from configparser import SafeConfigParser
from collections import defaultdict

"""
Configuration functions
"""


def read_config(file_config="dagon.ini", section=None):
    """
    Reads the configuration file specified

    :param file_config: path to the configuration file
    :type file_config: str

    :param section: section of the file to read
    :type section: str

    :return: dictionary with the configuration
    :rtype: dict(str, dict)
    """
    config = SafeConfigParser()
    config.read(file_config)
    if section is not None:
        try:
            return dict(config.items(section))
        except:
            return None
    else:
        dictionary = defaultdict(dict)
        for section in config.sections():
            dictionary[section] = {}
            for option in config.options(section):
                # print option
                dictionary[section][option] = config.get(section, option, raw=True)
        return dictionary



import socket


def is_port_open(host, port, timeout=5):
    """
    verifies if a port is open in a remote host

    :param host: IP of the remote host
    :type host: str

    :param port: port to check
    :type port: int

    :param timeout: timeout max to check
    :type timeout: int

    :return: True if the port is open
    :rtype: bool
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    result = sock.connect_ex((host, port))
    return result is 0



from fabric.api import local, env
from fabric.context_managers import settings, hide

from dagon.task import Task
from dagon.remote import RemoteTask
from subprocess import Popen, PIPE, STDOUT


class Batch(Task):
    """
    **Executes a Batch task**
    """

    def __init__(self, name, command, working_dir=None, globusendpoint=None, transversal_workflow=None):
        """
        :param name: task name
        :type name: str

        :param command: command to be executed
        :type command: str

        :param working_dir: path to the task's working directory
        :type working_dir: str

        :param globusendpoint: Globus endpoint ID
        :type globusendpoint: str
        """
        Task.__init__(self, name, command, working_dir,transversal_workflow = transversal_workflow, globusendpoint=globusendpoint)

    def __new__(cls, *args, **kwargs):
        """Create an Batch task local or remote

           Keyword arguments:
           name -- task name


           command -- command to be executed
           working_dir -- directory where the outputs will be placed
           ip -- hostname or ip of the machine where the task will be executed
           ssh_username -- username in remote machine
           keypath -- path to the private keypath
        """
        #if "ip" in kwargs:
        #    return super(Task, cls).__new__(RemoteBatch)
        #else:
        #    return super(Batch, cls).__new__(cls, *args, **kwargs)
        if "ip" in kwargs:
            return super().__new__(RemoteBatch)
        else:
            return super().__new__(cls)

    @staticmethod
    def execute_command(command):
        """
        Executes a local command

        :param command: command to be executed
        :type command: str
        :return: execution result
        :rtype: dict() with the execution output (str), code (int) and error (str)
        """
        # Execute the bash command
        # with settings(
        #         hide('warnings', 'running', 'stdout', 'stderr'),
        #         warn_only=True
        # ):
        #     result = local(command, capture=True)
        #     # check for an error
        #     code, message = 0, ""
        #     if len(result.stderr):
        #         code, message = 1, result.stderr
        #
        #     return {"code": code, "message": message, "output": result.stdout}
        p = Popen(command.split(" "), stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True, bufsize=-1, universal_newlines=True)
        #print "commmand",command
        out, err = p.communicate()

        code, message = 0, ""
        if len(err):
            code, message = 1, err
        return {"code": code, "message": message, "output": out}


    def on_execute(self, script, script_name):
        """
        Invoke the script specified

        :param script: content script
        :type script: str
        :param script_name: script name
        :type script_name: str
        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)
        """
        # Invoke the base method
        super(Batch, self).on_execute(script, script_name)
        return Batch.execute_command("bash " + self.working_dir + "/.dagon/" + script_name)

    # returns public key
    def get_public_key(self):
        """
        Return the temporal public key to this machine

        :return: public key
        :rtype: str with the public key
        """
        command = "cat " + self.working_dir + "/.dagon/ssh_key.pub"
        result = Batch.execute_command(command)
        return result['output']

    def add_public_key(self, key):
        """
        Add a SSH public key on the remote machine

        :param key: Path to the public key
        :type key: str
        :return: result of the execution
        :rtype: dict() with the execution output (str) and code (int)
        """
        command = "echo " + key.strip() + "| cat >> ~/.ssh/authorized_keys"
        result = Batch.execute_command(command)
        return result


class RemoteBatch(RemoteTask, Batch):
    """
    **Execute a Batch task on a remote machine**
    """

    def __init__(self, name, command, ssh_username=None, keypath=None, ip=None, working_dir=None, globusendpoint=None):
        """
        :param name: name of the task
        :type name: str

        :param command: command to be executed
        :type command: str

        :param ssh_username: UNIX username to connect through SSH
        :type ssh_username: str

        :param keypath: path to the public key
        :type keypath: str

        :param ip: IP address to connect to the remote machine
        :type ip: str

        :param working_dir: path of the working directory on the remote machine
        :type working_dir: str

        :param globusendpoint: Globus endpoint ID
        :type globusendpoint: str
        """
        RemoteTask.__init__(self, name, ssh_username, keypath, command, ip=ip, working_dir=working_dir,
                            globusendpoint=globusendpoint)

    def on_execute(self, launcher_script, script_name):
        """
        Execute a script on the remote machine

        :param script: script content
        :type script: str

        :param script_name: script name
        :type script_name: str

        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)
        """
        # Invoke the base method
        RemoteTask.on_execute(self, launcher_script, script_name)
        result = self.ssh_connection.execute_command("bash " + self.working_dir + "/.dagon/" + script_name)
        return result


class Slurm(Batch):
    """
    **Run a task using Slurm**

    :ivar partition: partition where the task will be executed
    :vartype partition: str

    :ivar ntask: number of parallel tasks to be executed
    :vartype ntask: int
    """

    def __init__(self, name, command, partition=None, ntasks=None, memory=None, working_dir=None, globusendpoint=None):

        """
        :param name: name of the task
        :type name: str

        :param command: command to be executed
        :type command: str

        :param partition: partition where the task will be executed
        :type partition: str

        :param ntasks: number of parallel tasks to be executed
        :type ntasks: int

        :param memory: number of memory to be allocate
        :type memory: int

        :param working_dir: path to the task's working directory
        :type working_dir: str

        :param globusendpoint: Globus endpoint ID
        :type globusendpoint: str
        """

        Batch.__init__(self, name, command, working_dir, globusendpoint=globusendpoint)
        self.partition = partition
        self.ntasks = ntasks
        self.memory = memory

    def __new__(cls, *args, **kwargs):
        """Create an Slurm task local or remote

           Keyword arguments:
           name -- task name
           command -- command to be executed
           partition -- partition where the task is going to be executed
           ntasks -- number of tasks to execute
           working_dir -- directory where the outputs will be placed
        """

        if "ip" in kwargs:
            return super().__new__(RemoteSlurm)
        else:
            return super().__new__(cls)

    def generate_command(self, script_name):

        """
        Generates the Slurm command including the partition and number of task parameters

        :param script_name: script to be executed
        :type script: str

        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)
        """

        partition_text = ""
        if self.partition is not None:
            partition_text = "--partition=" + self.partition

        ntasks_text = ""
        if self.ntasks is not None:
            ntasks_text = "--ntasks=" + str(self.ntasks)

        memory_text = ""
        if self.memory is not None:
            memory_text = "--mem=" + str(self.memory)

        # Add the slurm batch command
        # command = "sbatch " + partition_text + " " + ntasks_text + " --job-name=" + self.name + " --chdir=" + self.working_dir + " --output=" + self.working_dir + "/.dagon/stdout.txt --wait " + self.working_dir+"/.dagon/launcher.sh"
        command = "sbatch " + partition_text + " " + ntasks_text + " " + memory_text + " -J " + self.name + " -D " \
                  + self.working_dir + " -W " + self.working_dir + "/.dagon/" + script_name
        return command

    def on_execute(self, script, script_name):

        """
        Execute a script using slurm

        :param script: script content
        :type script: str

        :param script_name: script name
        :type script_name: str

        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)
        """

        super(Batch, self).on_execute(script, script_name)

        if script_name == "context.sh":
            return Batch.execute_command(self.working_dir + "/.dagon/" + script_name)

        command = self.generate_command(script_name)

        # Execute the bash command
        result = Batch.execute_command(command)
        return result


class RemoteSlurm(RemoteTask, Slurm):
    """
    ** Represent a task that runs on a remote slurm deployment **
    """

    def __init__(self, name, command, partition=None, ntasks=None, memory=None, working_dir=None, ssh_username=None, keypath=None,
                 ip=None, globusendpoint=None):
        """
        :param name: name of the task
        :type name: str

        :param command: command to be executed
        :type command: str

        :param partition: partition where the task will be executed
        :type partition: str

        :param ntasks: number of parallel tasks to be executed
        :type ntasks: int

        :param working_dir: path to the task's working directory
        :type working_dir: str

        :param globusendpoint: Globus endpoint ID
        :type globusendpoint: str

        :param ssh_username: UNIX username on the remote
        :type endpoint: str

        :param keypath: Path to the public key
        :type keypath: str

        :param ip: IP address to connect to the remote machine
        :type ip: str

        """
        Slurm.__init__(self, name, command, working_dir=working_dir, partition=partition, ntasks=ntasks, memory=memory,
                       globusendpoint=globusendpoint)
        RemoteTask.__init__(self, name, ssh_username, keypath, command, ip, working_dir, globusendpoint=None)

    def on_execute(self, script, script_name):
        """
        Execute a script using slurm

        :param script: script content
        :type script: str

        :param script_name: script name
        :type script_name: str

        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)
        """

        RemoteTask.on_execute(self, script, script_name)
        if script_name == "context.sh":
            return self.ssh_connection.execute_command("bash " + self.working_dir + "/.dagon/" + script_name)

        command = self.generate_command(script_name)
        # Execute the bash command
        result = self.ssh_connection.execute_command(command)
        return result





from requests.exceptions import ConnectionError
from requests.exceptions import MissingSchema
import logging
from ftplib import FTP
import os
import ftplib
import sys
# Perform the communication with the TPS manager
class FTP_API:
    def __init__(self, url, username="guess",password="guess"):
        self.base_url = url
        self.username = username
        self.password = password
        self.checkConnection()
        self.logger = logging.getLogger()


    # check if the service URL is valid or a service is available
    def checkConnection(self):
        """
        check if the service URL is valid or a service is available

        :raises ConnectionError: when it's not possible to connect to the URL provided
        """
        try:
            self.ftp=FTP(self.base_url)
            self.ftp.login(self.username,self.password)
        except ConnectionError as e:
            raise ConnectionError("It is not possible connect to the URL %s" % self.base_url)
        except MissingSchema:
            raise ConnectionError("Bad URL %s" % self.base_url)

    def downloadFiles(self,source,destination):
    #path & destination are str of the form "/dir/folder/something/"
    #path should be the abs path to the root FOLDER of the file tree to download
        try:
            self.ftp.cwd(source)
            #clone path to destination
            self.logger.error("destination : "+destination)

            os.makedirs(destination, exist_ok=True)
            os.chdir(destination) #workflow schema

        except OSError:
            #folder already exists at destination
            self.logger.error("folder exist")
            pass
        except ftplib.error_perm:
            #invalid entry (ensure input form: "/dir/folder/something/")
            print ("error: could not change to "+source)
            sys.exit("ending session")

        #list children:
        filelist=self.ftp.nlst()

        for file in filelist:
            try:
                #this will check if file is folder:
                self.ftp.cwd(file+"/")
                self.ftp.cwd("..")
                #if so, explore it:
                self.downloadFiles(file+"/",destination+"/"+file+"/")
            except ftplib.error_perm:

                #not a folder with accessible content
                #download & return
                os.chdir(destination)
                #possibly need a permission exception catch:
                self.ftp.retrbinary("RETR "+file, open(os.path.join(destination,file),"wb").write)
                print (file + " downloaded")
        self.ftp.cwd("..")

        return




import logging
import logging.config
import os
from logging.config import fileConfig
import threading
import copy 
from requests.exceptions import ConnectionError
from dagon.config import read_config

from time import time, sleep


class DAG_TPS(object):

    def __init__(self, name,config=None, config_file='dagon.ini', max_threads=10):
        """
        Create a meta-workflow

        :param name: Workflow name
        :type name: str

        """
        if config is not None:
            self.cfg = config
        else:
            self.cfg = read_config(config_file)
            fileConfig(config_file)

        self.workflows = []
        self.name = name
        self.logger = logging.getLogger()
        self.workflow_id = 0
        self.tasks = []
        self.T_tasks_needed = []
        self.T_tasks_needy = []
        self.is_api_available = False
        self.running = False





    def set_dry(self, dry):
        for wf in self.workflows:
            wf.dry = dry

    def add_workflow(self, workflow):
        """
        Add a workflow to this meta-workflow

        :param workflow: :class:`dagon.workflow` instance
        :type workflow: :class:`dagon.workflow`
        """
        self.workflows.append(workflow)
        workflow.set_dag_tps(self)
        #if self.is_api_available:
        #    self.api.add_task(self.workflow_id, task)

    def find_task_by_name(self, workflow_name, task_name):
        """
        Search for a task of an specific workflow

        :param workflow_name: Name of the workflow
        :type workflow_name: str

        :param task_name: Name of the task
        :type task_name: str

        :return: task instance
        :rtype: :class:`dagon.task.Task` instance if it is found, None in other case
        """

        # Check if the workflow is the current one
        for wf in self.workflows:
            if workflow_name == wf.name:
                wf.find_task_by_name(wf.name,task_name)
        return None

    def find_workflow_task(self,task_name):
        """
        Search for workflow name of an specific task

        :param task_name: Name of the task
        :type task_name: str

        :return: workflow name
        :rtype: string
        """
        # Check if the workflow is the current one
        for wf in self.workflows:
            if wf.find_task_by_name(wf.name,task_name) is not None:
                return wf.name
        return None


    def make_dependencies(self):
        """
        Looks for all the dependencies between workflows
        """
        for wf in self.workflows:
            #check the dependencies for each workflow
            wf.make_dependencies()
            #add the workflow's tasks to the DAGtps
        self.Validate_WF()
            

    # Return a json representation of the workflow
    def as_json(self,json_format="mw" ):
        """
        Return a json representation of the meta-workflow

        :param json_format: format of the json returned (mw o wf)
        :type json_format: str

        :return: JSON representation
        :rtype: dict(str, object) with data class
        """
        if json_format == "mw":
            jsonWorkflow = {"tasks": {}, "name": self.name, "id": self.workflow_id}
            for wf in self.workflows:
                for task in wf.tasks:
                    jsonWorkflow['tasks'][task.name] = task.as_json()
            return jsonWorkflow
        if json_format == "wf":
            jsonWorkflow = {"tasks": {}, "name": self.name, "id": self.workflow_id}
            for wf in self.workflows:
                for task in wf.tasks:
                    temp = task.remove_from_workflow()
                    jsonWorkflow['tasks'][task.name] = task.as_json()
                    jsonWorkflow['tasks'][task.name]['command'] = temp
            return jsonWorkflow


    def run(self):
        self.logger.debug("Running meta-workflow: %s", self.name)
        self.running=True
        start_time = time()
        #print self.tasks
        workflows= []
        for wf in self.workflows:
            workflows.append(threading.Thread(target=wf.run))
        for wf in workflows: wf.start()
        for wf in workflows: wf.join()

        completed_in = (time() - start_time)
        self.logger.info("Meta-Workflow '"+self.name+ "' completed in %s seconds ---" % completed_in)

    def Validate_WF(self):
        """
        Validate the transversality points to avoid any kind of cycle on the grap

        :param workflow: list of declared workflows
        :type workflow: list(class: dagon.workflow,...n)

        Raise an Exception when a cylce is founded

        self.tasks is fill with all the task of each workflow
        """
        for workflow in self.workflows:
            for task in workflow.tasks:
                for prev in task.prevs:
                    if workflow.find_task_by_name(workflow.name,prev) is None: #if is a task from another workflow -  'do you need someone?'
                        needed = False; needy = False
                        self.T_tasks_needed.append(prev) #dependency task is added to the transversal ones
                        #if the actual task is founded in the transversal, there exist a cycle.
                        if task in self.T_tasks_needed or task.nexts in self.T_tasks_needed: needed=True #are you or your decendents needed?
                        if prev in self.T_tasks_needy: needy=True #who you need is also needed?
                        if needy and needed:
                            logging.error('A cycle has been found')
                            raise Exception("A cycle has been found from %s(%s) to %s(%s)" % (prev.name, prev.workflow.name,task.name,task.workflow.name))
                        else:
                            self.T_tasks_needy.append(task) #add the task and decendets to the needys array
                            for t in task.nexts:
                                self.T_tasks_needy.append(t)
                temp = task.remove_from_workflow() #the command is changed, deleating the workflow reference
                self.tasks.append(temp)


"""
TPPs were depercated. Were used for analytics but are not useful anymore
"""


from os.path import abspath

from dagon.communication.ssh import SSHManager
from dagon.cloud import CloudManager
from dagon.task import Task


class RemoteTask(Task):
    """
    **Representates a task that is executed in a cloud instance, remote docker container,
    remote server or remove cluster HPC**

    :ivar ip: IP address to access to the remote machine
    :vartype ip: str

    :ivar keypath: Path to the SSH public key
    :vartype keypath: str

    :ivar ssh_username: Username to connect by SSH
    :vartype ssh_username: str

    :ivar ssh_connection: SSH connection with the machine
    :vartype ssh_connection: :class:`dagon.communDataFlow-Demication.ssh.SSHManager`
    """

    def __init__(self, name, ssh_username, keypath, command, ip=None, working_dir=None, globusendpoint=None,transversal_workflow=None):
        """

        :param name: Name of the task
        :type name: str

        :param ssh_username: UNIX username on the remote machine
        :type ssh_username: str

        :param keypath: Path to the public key
        :type keypath: str

        :param command: Command to be executed
        :type command: str

        :param ip: IP address to connect to the remote machine
        :type ip: str

        :param working_dir: Path of the working directory on the remote machine
        :type working_dir: str

        :param endpoint: Globus endpoint ID to transfer data with GridFTP
        :type endpoint: str

        """
        Task.__init__(self, name, command, working_dir=working_dir, transversal_workflow=transversal_workflow, globusendpoint=globusendpoint)
        self.transfer = None
        self.ip = ip
        self.keypath = abspath(keypath) if keypath is not None else keypath
        self.ssh_username = ssh_username
        self.ssh_connection = None
        #print name, self.ip, self.ssh_username
        if self.ip is not None and self.ssh_username is not None:
            self.ssh_connection = SSHManager(self.ssh_username, self.ip, self.keypath)

    def add_public_key(self, key):
        """
        Add a SSH public key on the remote machine

        :param key: Path to the public key
        :type key: str

        :return: result of the execution
        :rtype: dict() with result of the execution
        """
        command = "echo " + key.strip() + "| cat >> ~/.ssh/authorized_keys"
        result = self.ssh_connection.execute_command(command)
        return result

    def on_execute(self, script, script_name):
        """
        Execute an script on the remote machine

        :param script: script content
        :type key: str

        :param script_name: filename of the script
        :type key: str
        """
        # The launcher script name
        script_name = self.working_dir + "/.dagon/" + script_name
        # Create a temporary launcher script
        self.ssh_connection.create_file(script_name, script)

    # make dir
    def mkdir_working_dir(self, path):
        """
        Make a directory on the remote machine

        :param path: Path to the directory
        :type key: str

        :raises Exception: a problem occurred while the creation of the directory
        """
        #print self.ssh_connection
        res = self.ssh_connection.execute_command( "mkdir -p " + self.working_dir + "/.dagon")
        if res['code']:
            self.workflow.logger.error("%s: Error creating scratch directory on server %s", self.name, res['message'])
            raise Exception('Cannot create scratch directory on remote')

    # remove scratch directory
    def on_garbage(self):
        """
        Remove the scratch directory on the remote machine
        """
        self.ssh_connection.execute_command('mv {0} {1}'.format(self.working_dir,
                                            self.working_dir + "-removed"))

    def get_public_key(self):
        """
        Return the temporal public key to this machine

        :return: Public ke
        :rtype: str with the public key
        """
        command = "cat " + self.working_dir + "/.dagon/ssh_key.pub"
        result = self.ssh_connection.execute_command(command)
        return result['output']


class CloudTask(RemoteTask):
    """
    Class that represents a task executed on a cloud instance

    :ivar str instance_id: ID of the instance given by the cloud provider
    :ivar provider: cloud provider (see :class:`libcloud.compute.types.Provider`)
    :vartype provider: :class:`libcloud.compute.types.Provider`

    :ivar dict instance_flavour: characteristics of the instance
    :ivar dict key_options: ssh key options to connect to the instance
    :ivar node: Node where the task will be executed
    :vartype node: :class:`libcloud.compute.base.Node`

    :ivar bool stop_instance: True if the instance has to be stopped when the execution ends
    :ivar str instance_name: Name of the instance on the cloud provider
    """

    def __init__(self, name, command, provider, ssh_username, key_options, instance_id=None, instance_flavour=None,
                 instance_name=None, stop_instance=False, endpoint=None):
        """

        :param name: name of the task
        :type name: str

        :param command: command to be executed on the cloud instance
        :type command: str

        :param provider: cloud provider
        :type provider: :class:`libcloud.compute.types.Provider`

        :param ssh_username: UNIX username to connect through SSH
        :type ssh_username: str

        :param key_options: dictionary with the key options to connect to the instance
        :type key_options: dict

        :param instance_id: instance id on the cloud provider
        :type instance_id: str

        :param instance_flavour: characteristics of the instance (image, size, location)
        :type instance_flavour: dict

        :param instance_name: instance name on the cloud provider
        :type instance_name: str

        :param stop_instance: True if the instance has to be stoppend at the end of the task execution
        :type stop_instance: bool

        :param endpoint: Globus endpoint ID
        :type endpoint: str
        """
        super(CloudTask, self).__init__(name, ssh_username, key_options['key_path'], command, endpoint=endpoint)
        self.instance_id = instance_id
        self.provider = provider
        self.instance_flavour = instance_flavour
        self.key_options = key_options
        self.node = None
        self.stop_instance = stop_instance
        self.instance_name = instance_name

    def on_execute(self, script, script_name):
        """
        Execute a script on the cloud instance

        :param script: script content
        :type script: str

        :param script_name: script name
        :type script_name: str

        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)
        """

        RemoteTask.on_execute(self, script, script_name)
        return self.ssh_connection.execute_command("bash " + self.working_dir + "/.dagon/" + script_name)

    def execute(self):
        """
        Execute the task on the remote cloud instance.

        #. Create or get the instace from the cloud provider.
        #. Get the public IPS available.
        #. Open the SSH connection with the instace.
        #. Execute the task.

        :raises Exception: a problem occurred while the execution of the task
        """
        self.instance_name = self.instance_name if self.instance_name is not None else self.workflow.name.strip() + \
                                                                                       "-" + self.name
        self.node = CloudManager.get_instance(instance_id=self.instance_id, keyparams=self.key_options,
                                              flavour=self.instance_flavour, provider=self.provider,
                                              name=self.instance_name)
        self.ip = self.node.public_ips[0]
        self.ssh_connection = SSHManager(self.ssh_username, self.ip, self.keypath)
        super(CloudTask, self).execute()

    def decrement_reference_count(self):
        """
        Decremet the reference count. When the number of references is equals to zero, the garbage collector is called
        """
        self.reference_count = self.reference_count - 1

        # Check if the scratch directory must be removed
        if self.reference_count == 0 and self.stop_instance is True:
            # Call garbage collector (remove scratch directory, container, cloud instace, etc)
            self.ssh_connection.execute_command("shutdown -h now")
            # Perform some logging
            self.workflow.logger.debug("Removed instance %s", self.ip)

    def on_garbage(self):
        """
        Call the garbage collector

        #. Remove the scratch directory.
        #. The instance it is stopped if that is specified in the parameters of the task

        """
        RemoteTask.on_garbage(self)
        #if self.stop_instance:
        #    print("hola")
        #    self.ssh_connection.execute_command("shutdown -h now")



import logging
import logging.config
import os
from logging.config import fileConfig
import threading
import collections
from collections import abc

collections.MutableMapping = abc.MutableMapping
from backports.configparser import NoSectionError
from enum import Enum
from requests.exceptions import ConnectionError

from time import time, sleep

from dagon.config import read_config
from dagon.api import API
from dagon.api.server import WorkflowServer
from dagon.batch import Batch
from dagon.batch import Slurm
from dagon.communication.data_transfer import GlobusManager
from dagon.communication.data_transfer import SKYCDS
from dagon.docker_task import DockerRemoteTask
from dagon.remote import RemoteTask


class Status(Enum):
    """
    Possible states that a :class:`dagon.Task` could be in

    :cvar READY: Ready to execute
    :cvar WAITING: Waiting for some tasks to end them execution
    :cvar RUNNING: On execution
    :cvar FINISHED: Executed with success
    :cvar FAILED: Executed with error
    """

    READY = "READY"
    WAITING = "WAITING"
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    FAILED = "FAILED"


class Workflow(object):
    """
    **Represents a workflow executed by DagOn**

    :ivar name: unique name of the workflow
    :vartype name: str

    :ivar cfg: workflow configuration
    :vartype cfg: str

    :ivar tasks: Task to be executed by the workflow
    :vartype tasks: str

    :ivar workflow_id: workflow ID
    :vartype workflow_id: str

    :ivar is_api_available: True if the API is available
    :vartype is_api_available: str
    """

    SCHEMA = "workflow://"

    def __init__(self, name, config=None, config_file='dagon.ini', max_threads=10, jsonload=None):
        """
        Create a workflow

        :param name: Workflow name
        :type name: str

        :param config: configuration dictionary of the workflow
        :type config: dict(str, str)

        :param config_file: Path to the configuration file of the workflow. By default, try to loads 'dagon.ini'
        :type config_file: str
        """

        if config is not None:
            self.cfg = config
        else:
            self.cfg = read_config(config_file)
            fileConfig(config_file)
        self.sem = threading.Semaphore(max_threads)
        # supress some logs
        logging.getLogger("paramiko").setLevel(logging.WARNING)
        logging.getLogger("globus_sdk").setLevel(logging.WARNING)

        self.logger = logging.getLogger()
        self.dag_tps = None
        self.dry = False
        self.tasks = []
        self.workflow_id = 0
        self.is_api_available = False
        if jsonload is not None:  # load from json file
            self.load_json(jsonload)
        self.name = name

        # ftp attributes
        self.ftpAtt = dict()
        try:
            self.ftpAtt['host'] = self.cfg['ftp_pub']['ip']
            self.ftpAtt['user'] = "guess"
            self.ftpAtt['password'] = "guess"
            self.local_path = self.cfg['batch']['scratch_dir_base']
        except KeyError:
            self.logger.error("No ftp ip in config file")

        # to regist in the dagon service
        if self.cfg['dagon_service']['use'] == "True":
            try:
                #self.logger.debug("verifing dagon service")
                self.api = API(self.cfg['dagon_service']['route'])
                self.is_api_available = True
            except KeyError:
                self.logger.error("No dagon URL in config file")
            except NoSectionError:
                self.logger.error("No dagon URL in config file")
            except ConnectionError as e:
                self.logger.error(e)

        if self.is_api_available:
            try:
                self.workflow_id = self.api.create_workflow(self)
                self.logger.debug("Workflow registration success id = %s" % self.workflow_id)
            except Exception as e:
                raise Exception(e)

        self.data_mover = DataMover.COPY
        self.stager_mover = StagerMover.NORMAL

    def get_dry(self):
        return self.dry

    def set_dry(self, dry):
        self.dry = dry

    def get_data_mover(self):
        return self.data_mover

    def set_data_mover(self, data_mover):
        self.data_mover = data_mover

    def set_stager_mover(self, stager_mover):
        self.stager_mover = stager_mover

    def get_scratch_dir_base(self):
        """
        Returns the path to the base scratch directory

        :return: Absolute path to the scratch directory
        :rtype: str with absolute path to the base scratch directory
        """
        return self.cfg['batch']['scratch_dir_base']

    def find_task_by_name(self, workflow_name, task_name):
        """
        Search for a task of an specific workflow

        :param workflow_name: Name of the workflow
        :type workflow_name: str

        :param task_name: Name of the task
        :type task_name: str

        :return: task instance
        :rtype: :class:`dagon.task.Task` instance if it is found, None in other case
        """

        # Check if the workflow is the current one
        if workflow_name == self.name:

            # For each task in the tasks collection
            for task in self.tasks:

                # Check the task name
                if task_name == task.name:
                    # Return the result
                    return task

        return None

    def add_task(self, task):
        """
        Add a task to this workflow

        :param task: :class:`dagon.task.Task` instance
        :type task: :class:`dagon.task.Task`
        """
        if task.data_mover is None:
            task.set_data_mover(self.data_mover)
        if task.stager_mover is None:
            task.set_stager_mover(self.stager_mover)

        self.tasks.append(task)
        task.set_workflow(self)
        if self.is_api_available:
            self.api.add_task(self.workflow_id, task)

    def set_dag_tps(self, DAG_tps):
        """
        Set the DAG_tps workflow which execute this workflow

        :param  DAG_tps: :class:`dagon.dag_tps` instance
        :type  DAG_tps: :class:`dagon.dag_tps`
        """
        self.dag_tps = DAG_tps

    def make_dependencies(self):
        """
        Looks for all the dependencies between tasks
        """

        # Clean all dependencies
        for task in self.tasks:
            task.nexts = []
            task.prevs = []
            task.reference_count = 0

        # Automatically detect dependencies
        for task in self.tasks:
            # Invoke pre run
            task.set_semaphore(self.sem)
            task.set_dag_tps(self.dag_tps)
            task.pre_run()
        self.Validate_WF()

    # Return a json representation of the workflow
    def as_json(self):
        """
        Return a json representation of the workflow

        :return: JSON representation
        :rtype: dict(str, object) with data class
        """

        jsonWorkflow = {"tasks": {}, "name": self.name, "id": self.workflow_id, "host": self.ftpAtt["host"]}
        for task in self.tasks:
            jsonWorkflow['tasks'][task.name] = task.as_json()
        return jsonWorkflow

    def run(self):
        self.logger.debug("Running workflow: %s", self.name)
        start_time = time()
        for task in self.tasks:
            try:
                task.start()
            except:
                pass

        for task in self.tasks:
            try:
                task.join()
            except:
                pass
        
        completed_in = (time() - start_time)
        self.logger.info("Workflow '" + self.name + "' completed in %s seconds ---" % completed_in)

    def load_json(self, Json_data):
        from dagon.task import DagonTask, TaskType
        self.name = Json_data['name']
        self.workflow_id = Json_data['id']
        for task in Json_data['tasks']:
            temp = Json_data['tasks'][task]
            tk = DagonTask(TaskType[temp['type'].upper()], temp['name'], temp['command'])
            self.add_task(tk)
        # self.make_dependencies()

    def Validate_WF(self):
        """
        Validate the workflow to avoid any kind of cycle on the grap

        Raise an Exception when a cylce is founded
        """
        needed = []
        needy = []
        for task in self.tasks:
            for prev in task.prevs:
                bool_needed = False
                bool_needy = False
                needed.append(prev)  # dependency task is added
                if task in needed or task.nexts in needed: bool_needed = True  # are you or your decendents needed?
                if prev in needy: bool_needy = True  # that who you need is also needed?
                if bool_needy and bool_needed:
                    logging.warning('A cycle have been found')
                    raise Exception("A cycle have been found from %s to %s" % (prev.name, task.name))
                else:
                    needy.append(task)  # add the task and decendets to the needys array
                    for t in task.nexts:
                        needy.append(t)


class DataMover(Enum):
    """
    Possible transfer protocols/apps

    :cvar DONTMOVE: Don't move nothing
    :cvar LINK: Using a symbolic link
    :cvar COPY: Copying the data
    :cvar SCP: Using secure copy
    :cvar HTTP: Using HTTP
    :cvar HTTPS: Using HTTPS
    :cvar FTP: Using FTP
    :cvar SFTP: Using secure FTP
    :cvar GRIDFTP: Using Globus GridFTP
    """

    DONTMOVE = 0
    LINK = 1
    COPY = 2
    SCP = 3
    HTTP = 4
    HTTPS = 5
    FTP = 6
    SFTP = 7
    GRIDFTP = 8
    SKYCDS = 9


class StagerMover(Enum):
    """
    Possible mode

    :cvar NORMAL: sequential
    :cvar PARALLEL: using threads
    :cvar SLURM: using Slurm
    """
    NORMAL = 0
    PARALLEL = 1
    SLURM = 2


class ProtocolStatus(Enum):
    """
    Status of the protocol on a machine

    :cvar ACTIVE: Protocol running and active
    :cvar NONE: Protocol not installed
    :cvar INACTIVE: Protocol not running
    """

    ACTIVE = "active"
    NONE = "none"
    INACTIVE = "inactive"


class Stager(object):
    """
    Choose the transference protocol to move data between tasks
    """

    def __init__(self, data_mover, stager_mover, cfg):
        self.data_mover = data_mover
        self.stager_mover = stager_mover
        self.cfg = cfg

    def stage_in(self, dst_task, src_task, dst_path, local_path):
        """
        Evaluates the context of the machines and choose the transfer protocol

        :param dst_task: task where the data has to be put
        :type dst_task: :class:`dagon.task.Task`

        :param src_task: task from the data has to be taken
        :type src_task: :class:`dagon.task.Task`

        :param dst_path: path where the file is going to be save on the destiny directory
        :type dst_path: str

        :param local_path: path where is the file to be transferred on the source task
        :type local_path: str

        :return: comand to move the data
        :rtype: str with the command
        """

        data_mover = DataMover.DONTMOVE
        command = ""

        # ToDo: this have to be make automatic
        # get tasks info and select transference protocol
        dst_task_info = dst_task.get_info()
        src_task_info = src_task.get_info()


        # check transference protocols and remote machine info if is available
        if dst_task_info is not None and src_task_info is not None:
            if dst_task_info['ip'] == src_task_info['ip']:
                data_mover = self.data_mover
            else:
                protocols = ["GRIDFTP", "SCP", "FTP"]
                for p in protocols:
                    if ProtocolStatus(src_task_info[p]) is ProtocolStatus.ACTIVE:
                        data_mover = DataMover[p]

                        if data_mover == DataMover.GRIDFTP and \
                                not dst_task.get_endpoint() and \
                                not src_task.get_endpoint():
                            continue

                        break
        else:  # best effort (SCP)
            data_mover = self.data_mover

        src = src_task.get_scratch_dir() + "/" + local_path

        dst = dst_path + "/" + os.path.dirname(os.path.abspath(local_path))

        # Check if the symbolic link have to be used...
        if data_mover == DataMover.GRIDFTP:
            # data could be copy using globus sdk
            ep1 = src_task.get_endpoint()
            ep2 = dst_task.get_endpoint()
            gm = GlobusManager(ep1, ep2, self.cfg["globus"]["clientid"], self.cfg["globus"]["intermadiate_endpoint"])

            # generate tar with data
            #tar_path = src + "/data.tar"
            #command_tar = "tar -czvf %s %s --exclude=*.tar" % (tar_path, src_task.get_scratch_dir())
            #result = src_task.execute_command(command_tar)

            #get filename from path 
            intermediate_filename = os.path.basename(local_path)
            dst = dst_path + "/" + os.path.dirname(os.path.abspath(local_path)) + "/" + intermediate_filename

            gm.copy_data(src, dst, intermediate_filename)# + "/" + "data.tar.gz")

        elif data_mover == DataMover.SKYCDS:
            skycds = SKYCDS()
            upload_result = skycds.upload_data(src_task, src_task.get_scratch_dir(), encryption=True)
            download_result = skycds.download_data(dst_task, dst_path)


        elif data_mover == DataMover.LINK:
            # Add the link command
            command = command + "# Add the link command\n"
            cmd = "ls $file $dst"
            if StagerMover(self.stager_mover) == StagerMover.PARALLEL:
                cmd = "ls {} $dst"
            command = command + self.generate_command(src, dst, cmd, self.stager_mover.value)
                      
        # Check if the copy have to be used...
        elif data_mover == DataMover.COPY:
            # Add the copy command
            command = command + "# Add the copy command\n"
            cmd = "cp -r $file $dst"
            if StagerMover(self.stager_mover) == StagerMover.PARALLEL:
                cmd = "cp -r {} $dst"
            command = command + self.generate_command(src, dst, cmd, self.stager_mover.value)

        # Check if the secure copy have to be used...
        elif data_mover == DataMover.SCP:
            # Add the copy command
            command = command + "# Add the secure copy command\n"
            if isinstance(src_task, RemoteTask):  # if source is accessible from destiny machine
                # copy my public key
                key = dst_task.get_public_key()
                src_task.add_public_key(key)
                cmd = "scp -r -o LogLevel=ERROR -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -i " + dst_task.working_dir + \
                          "/.dagon/ssh_key -r " + src_task.get_user() + "@" + src_task.get_ip() + ":" + \
                           "$file $dst \n\n"
                if StagerMover(self.stager_mover) == StagerMover.PARALLEL:
                    cmd = "scp -r -o LogLevel=ERROR -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -i " + dst_task.working_dir + \
                          "/.dagon/ssh_key -r " + src_task.get_user() + "@" + src_task.get_ip() + ":" + \
                           "{} $dst \n\n"
                command = command + self.generate_command(src, dst, cmd, self.stager_mover.value)
                command += "\nif [ $? -ne 0 ]; then code=1; fi"
                # command += "\n rm " + dst_task.working_dir + "/.dagon/ssh_key"
            else:  # if source is a local machine
                # copy my public key
                key = src_task.get_public_key()
                dst_task.add_public_key(key)

                command_mkdir = "mkdir -p " + dst_path + "/" + os.path.dirname(local_path) + "\n\n"
                res = dst_task.ssh_connection.execute_command(command_mkdir)

                if res['code']:
                    raise Exception("Couldn't create directory %s" % dst_path + "/" + os.path.dirname(local_path))

                cmd = "scp -r -o LogLevel=ERROR -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -i " + src_task.working_dir + \
                                "/.dagon/ssh_key -r " + " $file " + \
                                dst_task.get_user() + "@" + dst_task.get_ip() + ":$dst \n\n"
                if StagerMover(self.stager_mover) == StagerMover.PARALLEL:
                    cmd = "scp -r -o LogLevel=ERROR -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -i " + src_task.working_dir + \
                                "/.dagon/ssh_key -r " + " {} " + \
                                dst_task.get_user() + "@" + dst_task.get_ip() + ":$dst \n\n"
                command_local = self.generate_command(src, dst, cmd, self.stager_mover.value)
                res = Batch.execute_command(command_local)

                if res['code']:
                    raise Exception("Couldn't copy data from %s to %s" % (src_task.get_ip(), dst_task.get_ip()))

        command += "\nif [ $? -ne 0 ]; then code=1; fi"

        return command

    def generate_command(self, src, dst, cmd, mode):
        return """
#! /bin/bash

src={}
dst={}
mode={}
jobs={}
partition={}

for file in $src
do
cmd="{}"
case $mode in
    1)
    # Run in parallel using local queue
    find $src -type f,l | parallel -j$jobs "$cmd"
    break
    ;;
    2)
    # Run in parallel using slurm
    srun --partition=$partition --ntasks=1 --cpus-per-task=1 $cmd &
    ;;
    *)
    # Run requentially
    $cmd
    ;;
esac
done

wait
        """.format(src, dst, mode, self.cfg["batch"]["threads"], self.cfg["sulrm"]["partition"], cmd)



from fabric.api import local, settings, hide


class DockerClient(object):
    """
    Manages the communication with a docker container
    """

    def exec_command(self, command):
        """
        executes a command on a container

        :param command: command to be executed
        :type command: str

        :return: execution results
        :rtype: dict(str, object)
        """

        with settings(
                hide('warnings', 'running', 'stdout', 'stderr'),
                warn_only=True
        ):
            res = local(command, capture=True)

            if len(res.stderr):
                return {"code": 1, "message": res.stderr}
            elif res.return_code != 0:
                return {"code": 1, "message": res.stdout}
            else:
                return {"code": 0, "output": res.stdout}

    @staticmethod
    def form_string_cont_creation(image, command=None, volume=None, dagon_volume=None, ports=None):
        """
        creates an string to create the container on the remote machine

        :param image: image of the container
        :type image: str

        :param command: command to be executed
        :type command: str

        :param volume: shared volume with the host system
        :type volume: dict(str, str) => {"host":"host_path", "container":"container_path"}

        :param ports: ports expose with the host system
        :type ports: dict(str, str) => {"host":"host_port", "container":"container_port"}

        :return: string with the command to create the container
        :rtype: str
        """

        docker_command = "docker run --net=\"host\""

        docker_command += " -t -d"

        if volume is not None:
            docker_command += " -v \'%s\':\'%s\'" % (volume['host'], volume['container'])
        if dagon_volume is not None:
            docker_command += " -v \'%s\':\'%s\'" % (dagon_volume['host'], dagon_volume['container'])
        if ports is not None:
            docker_command += " -p \'%s\':\'%s\'" % (ports['host'], ports['container'])
        docker_command += " %s" % image

        if command is not None:
            docker_command += " " + command

        return docker_command


class DockerRemoteClient(DockerClient):

    """
    Manages the execution of command on remote containers
    """

    def __init__(self, ssh):
        """
        :param ssh: ssh connection with remote host
        """
        self.ssh = ssh

    def exec_command(self, command):
        """
        executes a command on remote container

        :param command: command to execute on the remote machine
        :type command: str

        :return: execution results
        :rtype: dict(str, object)
        """
        with settings(
                hide('warnings', 'running', 'stdout', 'stderr'),
                warn_only=True
        ):
            result = self.ssh.execute_command(command)
            return result


class Container(object):

    """
    **Represents a docker container**
    """

    def __init__(self, key, client):
        """
        :param key: docker container key
        :type key: str

        :param client: docker client
        :type client: :class:`DockerClient`
        """
        self.key = key
        self.client = client

    def logs(self):
        """
        get the logs from the container

        :return: container logs
        :rtype: str
        """
        command = "docker logs " + self.key
        res = self.client.exec_command(command)
        return res

    def exec_in_cont(self, command):
        """
        create an string with the command to execute something on the container

        :param command: command to execute
        :type command: str

        :return: container command
        :rtype: str
        """
        docker_command = "docker exec -t " + self.key + " sh -c \"" + command.strip() + "\""
        # res = self.client.exec_command(docker_command)
        return docker_command

    def rm(self, force=False):
        """
        removes a docker container

        :param force: force to remove a container in execution
        :rtype force: bool

        :return: True if the container has been removed
        :rtype: bool
        """

        band = "-f" if force else ""
        command = "docker rm %s %s" % (band, self.key)

        try:
            self.client.exec_command(command)
        except Exception as e:
            return False
        return True

    def stop(self):
        """
        stops a container in execution

        :return: True if the container has been stopped
        :rtype: bool
        """
        command = "docker stop %s" % self.key
        try:
            self.client.exec_command(command)
        except Exception as e:
            print(e)
            return False
        return True



import os
import time

from enum import Enum
from libcloud.compute.providers import get_driver
from libcloud.compute.types import NodeState
from libcloud.compute.types import Provider

from dagon.config import read_config


class CloudManager(object):
    """
    Manages the connection with the cloud providers using LibCloud
    """

    @staticmethod
    def get_instance(keyparams, provider, name=None, flavour=None, instance_id=None):
        """
        Get or create an instance in the specified cloud provider

        :param keyparams: dictionary with the ssh key
        :type keyparams: dict(str, str)

        :param provider: cloud provider
        :type provider: :class:`libcloud.compute.types.Provider`

        :param name: instance name (for existing instances)
        :type name: str

        :param flavour: dictionary with the instance configuration (size, region, operating system)
        :type flavour: dict

        :param instance_id: instance id in the cloud provider (for existing instances)
        :type instance_id: str

        :return: instance created
        :rtype: :class:`libcloud.compute.base.Node`
        """

        driver = get_driver(provider)
        conf = read_config(section=str(provider))
        if provider == Provider.GCE:
            conn = driver(conf['key'], conf['secret'], project=conf['project'])
        else:
            conn = driver(**conf)
        manager = globals()[str(provider).upper()]
        node = manager.createInstance(conn, name, flavour, keyparams) if flavour is not None \
            else CloudManager.get_existing_instance(conn, id=instance_id, name=name)
        
        node = CloudManager.wait_until_running(conn, node)
        return node

    @staticmethod
    def wait_until_running(conn, node):
        """
        wait until an instance is running

        :param conn: driver to the cloud provider
        :type conn: Libcloud driver

        :param node: node to waite
        :type node: :class:`libcloud.compute.base.Node`

        :return: instance updated
        :rtype node: :class:`libcloud.compute.base.Node`
        """

        while node.state is not NodeState.RUNNING:
            try:
                node = CloudManager.get_existing_instance(conn, uuid=node.uuid)
            except Exception:
                pass
            time.sleep(1)
        return node

    @staticmethod
    def create_instance(conn, name, flavour, keyparams):
        """
        create a new instance on the cloud provider

        :param conn: driver to the cloud provider
        :type conn: Libcloud driver

        :param name: instance name
        :param name: str

        :param flavour: dictionary with the instance configuration (size, region, operating system)
        :type flavour: dict

        :param keyparams: dictionary with the ssh key
        :type keyparams: dict(str, str)

        :return: instance created
        :rtype node: :class:`libcloud.compute.base.Node`
        """

        if flavour is None:
            raise Exception('The characteristics of the image has not been specified')

        sizes = conn.list_sizes()
        size = [s for s in sizes if s.id == flavour['size']]
        image = conn.get_image(flavour['image'])
        size = size[0] if len(size) > 0 else None
        key = None

        if image is None or size is None:
            raise Exception('Size or image doesn\'t exists')

        if keyparams['option'] == KeyOptions.CREATE:
            key = KeyPair.createPairKey(conn, keyparams['keypath'], keyparams['cloudargs'])
        elif keyparams['option'] == KeyOptions.GET:
            key = KeyPair.getExistingPairKey(conn, keyparams['keyname'])
        elif keyparams['option'] == KeyOptions.IMPORT:
            key = KeyPair.importKey(conn, keyparams['keypath'])

        node = conn.create_node(name=name, image=image, size=size,
                                ex_keyname=key.name)
        return node

    @staticmethod
    def get_existing_instance(conn, id=None, name=None, uuid=None):
        """
        return an existing instance from the cloud provider. The id, name or uuid is used to get the instance.


        :param conn: driver to the cloud provider
        :type conn: Libcloud driver

        :param id: instance id
        :type id: str

        :param name: instance name
        :type name: str

        :param uuid: instance uuid
        :type uuid: str

        :return: instance
        :rtype node: :class:`libcloud.compute.base.Node`
        """
        if id is None and name is None and uuid is None:
            raise Exception('Must specified an intance\'s id or name')
        nodes = conn.list_nodes()
        node = None
        if id is not None:
            node = [node for node in nodes if node.id == id]
        elif name is not None:
            node = [node for node in nodes if node.name == name]
        elif uuid is not None:
            node = [node for node in nodes if node.uuid == uuid]
        if len(node) == 0:
            raise Exception('Instance doesn\'t exists')
        return node[0]


class KeyOptions(Enum):
    """
    Represents the different options to manage the keys instance. The key could be create from the scratch,
    import an existing local key or get an existing key on the cloud provider

    :cvar CREATE: create a new key
    :cvar IMPORT: import a key
    :cvar GET: get an existing key
    """

    CREATE = "CREATE"
    IMPORT = "IMPORT"
    GET = "GET"


class KeyPair(object):

    @staticmethod
    def generate_RSA(bits=2048):
        """
        Generate an RSA keypair with an exponent of 65537 in PEM format

        :param bits: bits The key length in bits
        :type bits: int

        :return: Return private key and public key
        """

        from Crypto.PublicKey import RSA
        new_key = RSA.generate(bits, e=65537)
        public_key = new_key.publickey().exportKey("OpenSSH")
        private_key = new_key.exportKey("PEM")
        return private_key, public_key

    @staticmethod
    def writeKey(privateKey, filename):
        """
        writes the private key in a file

        :param privateKey: key string
        :type privateKey: str

        :param filename: path to the file where the key will be saved
        :type filename: str
        """

        from os import chmod
        with open(filename, 'w') as content_file:
            chmod(filename, 0o600)
            content_file.write(privateKey)

    @staticmethod
    def createPairKey(conn, filename, args):
        """
        create a new key on the cloud provider

        :param conn: driver to the cloud provider
        :type conn: Libcloud driver

        :param filename: path to the file where the key will be saved
        :type filename: str

        :param args: arguments to create the key
        :type args: dict()

        :return: key pair object
        """

        from inspect import getargspec

        # CHECK FOR THE PARAMS OF THE FUNCTION
        sig = getargspec(conn.create_key_pair)
        foo_args = sig.args
        foo_params = dict()

        for arg in foo_args:
            if arg in args:
                foo_params[arg] = args[arg]

        # CREATE NEW KEY
        key_pair = conn.create_key_pair(**foo_params)
        private_key = key_pair.private_key
        if private_key is None:
            private_key = args['private_key']
        KeyPair.writeKey(private_key, filename)
        return key_pair

    @staticmethod
    def getExistingPairKey(conn, keyname):
        """
        returns an existing pair key from the cloud provider

        :param conn: driver to the cloud provider
        :type conn: Libcloud driver

        :param keyname: name of the key on the cloud provider
        :type keyname: str

        :return: key pair object
        """
        keys = conn.list_key_pairs()
        key_pair = [key for key in keys if key.name == keyname]
        key_pair = key_pair[0] if len(key_pair) > 0 else None
        return key_pair

    @staticmethod
    def importKey(conn, key_path):
        """
        import an existing key from a local file

        :param conn: driver to the cloud provider
        :type conn: Libcloud driver

        :param key_path: path to the key
        :param key_path: str

        :return: key pair object
        """

        key_file_path = os.path.expanduser(key_path)
        key_pair = conn.import_key_pair_from_file(name=key_path,
                                                  key_file_path=key_file_path)
        return key_pair


class EC2(object):
    """
    Manages the creation of instances on Amazon EC2 using libcloud
    """

    @staticmethod
    def createInstance(conn, name, flavour, keyparams):
        """
        create a new instance on amazon EC2

        :param conn: driver to the cloud provider
        :type conn: EC2 Libcloud driver

        :param name: instance name
        :param name: str

        :param flavour: dictionary with the instance configuration (size, region, operating system)
        :type flavour: dict

        :param keyparams: dictionary with the ssh key
        :type keyparams: dict(str, str)

        :return: instance created
        :rtype node: :class:`libcloud.compute.base.Node`
        """

        if flavour is None:
            raise Exception('The characteristics of the image has not been specified')

        sizes = conn.list_sizes()
        size = [s for s in sizes if s.id == flavour['size']]
        image = conn.get_image(flavour['image'])
        size = size[0] if len(size) > 0 else None
        key = None

        if image is None or size is None:
            raise Exception('Size or image doesn\'t exists')
        if keyparams['option'] == KeyOptions.CREATE:
            key = KeyPair.createPairKey(conn, keyparams['key_path'], keyparams['cloud_args'])
        elif keyparams['option'] == KeyOptions.GET:
            key = KeyPair.getExistingPairKey(conn, keyparams['cloud_args']['name'])
        elif keyparams['option'] == KeyOptions.IMPORT:
            key = KeyPair.importKey(conn, keyparams['key_path'])

        node = conn.create_node(name=name, image=image, size=size,
                                ex_keyname=key.name)
        return node


class DIGITALOCEAN(object):
    """
    Manages the creation of instances on DigitalOcean using libcloud
    """

    @staticmethod
    def createInstance(conn, name, flavour, keyparams):
        """
        create a new instance on DigitalOcean

        :param conn: driver to the cloud provider
        :type conn: DigitalOcean Libcloud driver

        :param name: instance name
        :param name: str

        :param flavour: dictionary with the instance configuration (size, region, operating system)
        :type flavour: dict

        :param keyparams: dictionary with the ssh key
        :type keyparams: dict(str, str)

        :return: instance created
        :rtype node: :class:`libcloud.compute.base.Node`
        """

        if flavour is None:
            raise Exception('The characteristics of the image has not been specified')

        sizes = conn.list_sizes()
        size = [s for s in sizes if s.id == flavour['size']]
        image = conn.get_image(flavour['image'])
        size = size[0] if len(size) > 0 else None
        locations = conn.list_locations()
        location = [l for l in locations if l.id == flavour['location']]
        location = location[0] if len(location) > 0 else None
        key = None

        if image is None or size is None or location is None:
            raise Exception('Size, location or image doesn\'t exists')
        if keyparams['option'] == KeyOptions.CREATE:
            key = KeyPair.createPairKey(conn, keyparams['key_path'], keyparams['cloudargs'])
        elif keyparams['option'] == KeyOptions.GET:
            key = KeyPair.getExistingPairKey(conn, keyparams['keyname'])
        elif keyparams['option'] == KeyOptions.IMPORT:
            key = KeyPair.importKey(conn, keyparams['keypath'])

        node = conn.create_node(name=name, image=image, size=size, location=location,
                                ex_create_attr={"ssh_keys": [key.fingerprint]})
        return node


class GCE(object):
    """
    Manages the creation of instances on google cloud using libcloud
    """

    @staticmethod
    def createInstance(conn, name, flavour, keyparams):
        """
        create a new instance on google cloud

        :param conn: driver to the cloud provider
        :type conn: google cloud Libcloud driver

        :param name: instance name
        :param name: str

        :param flavour: dictionary with the instance configuration (size, region, operating system)
        :type flavour: dict

        :param keyparams: dictionary with the ssh key
        :type keyparams: dict(str, str)

        :return: instance created
        :rtype node: :class:`libcloud.compute.base.Node`
        """

        if flavour is None:
            raise Exception('The characteristics of the image has not been specified')

        image = flavour['image']
        location = flavour['location']
        size = flavour['size']

        metadata = {"items": [
            {"value": "%s: %s %s" % (keyparams['username'], keyparams['public_key'], keyparams['username']),
             "key": "ssh-keys"}]}

        KeyPair.writeKey(keyparams["private_key"], keyparams['keypath'])
        node = conn.create_node(name=name, image=image, size=size, location=location, ex_metadata=metadata)

        return node



from dagon import Batch
from dagon.remote import RemoteTask
from dagon.dockercontainer import Container
from dagon.dockercontainer import DockerClient
from dagon.dockercontainer import DockerRemoteClient
from dagon.task import Task

import docker


class DockerTask(Batch):
    """
    ***Represents a task running on a docker container***

    :ivar docker_client: client to manages the container
    :vartype docker_client: :class:`dagon.dockercontainer.DockerClient`

    :ivar container: represents a docker container
    :vartype container: :class:`dagon.dockercontainer.Container`

    """

    def __init__(self, name, command, image=None, container_id=None, working_dir=None, globusendpoint=None, remove=True, volume=None,transversal_workflow=None):

        """
        :param name: task name
        :type name: str

        :param command: command to be executed
        :type command: str

        :param working_dir: path to the task's working directory
        :type working_dir: str

        :param image: container image
        :type image: str

        :param globusendpoint: Globus endpoint ID
        :type globusendpoint: str

        :param remove: if it's True the container is removed after the task ends its execution
        :type remove: bool
        """

        Task.__init__(self, name, command, working_dir=working_dir,transversal_workflow=transversal_workflow, globusendpoint=globusendpoint)
        self.command = command
        self.container_id = container_id
        self.container = None
        self.remove = remove
        self.image = image
        self.volume = volume
        self.docker_client = DockerClient()

    def __new__(cls, *args, **kwargs):
        if "ip" in kwargs:
            return super(Task, cls).__new__(DockerRemoteTask)
        else:
            return super(DockerTask, cls).__new__(cls)

    def include_command(self, body):
        """
        Include the command to execute in the script body

        :param body: Script body
        :type body: str

        :return: Script body with the command
        :rtype: string
        """

        body = super(DockerTask, self).include_command(body)
        body = "cd " + self.working_dir + ";" + body
        body = self.container.exec_in_cont(body) + "\n"
        return body

    def pre_process_command(self, command):
        """
        Add some post process commands after the task execution. Also creates the docker container.

        :param command: Command to be executed
        :type command: str

        :return: Command post-processed
        :rtype: string
        """

        if self.container is None:
            self.container_id = self.create_container() if self.container_id is None else self.container_id
            self.container = Container(self.container_id.rstrip(), self.docker_client)
        return super(DockerTask, self).pre_process_command(command)

    def pull_image(self, image):
        """
        Pull a Docker image from Docker Hub

        :param image: Image name
        :type image: str

        :return: pull result
        :rtype: dict()
        """

        client = docker.from_env()  # Connect to the Docker daemon

        try:
            client.images.pull(image)  # Pull the Docker image
            self.workflow.logger.info("%s: Successfully pulled %s", self.name, image)
        except docker.errors.ImageNotFound:
            self.workflow.logger.error("%s: Image %s not found", self.name, image)
        except docker.errors.APIError as e:
            print(f"An error occurred: {e}")

        #return self.docker_client.pull_image(image)

    # Create a Docker container
    def create_container(self):
        """
        Creates the container where the task will be executed

        :return: container key
        :rtype: string

        :raises Exception: a problem occurred while container creation
        """

        self.pull_image(self.image) #pull image


        command = DockerClient.form_string_cont_creation(image=self.image,
                                                         volume=self.volume,
                                                         dagon_volume={"host": self.workflow.get_scratch_dir_base(),
                                                                 "container": self.workflow.get_scratch_dir_base()})
       
        result = self.docker_client.exec_command(command)
        if result['code']:
            raise Exception(self.result["message"].rstrip())
        return result['output']

    def remove_container(self):
        """
        Removes a docker container
        """
        self.container.stop()
        if self.remove:
            self.container.rm()

    def on_execute(self, script, script_name):
        """
        Execute the task script

        :param script: script content
        :type script: str

        :param script_name: script name
        :type script_name: str

        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)

        """

        # Invoke the base method
        Task.on_execute(self, script, script_name)
        return Batch.execute_command("bash " + self.working_dir + "/.dagon/" + script_name)
        # return self.docker_client.exec_command(self.working_dir + "/.dagon/" + script_name)"""

    def on_garbage(self):
        """
        Call garbage collector, removing the scratch directory, containers and instances related to the
        task
        """
        super(DockerTask, self).on_garbage()
        self.remove_container()


class DockerRemoteTask(RemoteTask, DockerTask):
    """
    **Represents a docker task running on a remote machine**

    :ivar docker_client: client to manages the container
    :vartype docker_client: :class:`dagon.dockercontainer.DockerRemoteClient`
    """

    def __init__(self, name, command, image=None, container_id=None, ip=None, ssh_username=None, keypath=None,
                 working_dir=None, remove=True, globusendpoint=None):
        """
        :param name: task name
        :type name: str

        :param command: command to be executed
        :type command: str

        :param working_dir: path to the task's working directory
        :type working_dir: str

        :param image: container image
        :type image: str

        :param globusendpoint: Globus endpoint ID
        :type globusendpoint: str

        :param remove: if it's True the container is removed after the task ends its execution
        :type remove: bool

        :param ip: IP address of the machine where the container will be created
        :type ip: str

        :param ssh_username: UNIX username to connect through SSH
        :type ssh_username: str

        :param keypath: Path to the public key
        :type keypath: str
        """

        DockerTask.__init__(self, name, command, container_id=container_id, working_dir=working_dir, image=image,
                            remove=remove, globusendpoint=globusendpoint)
        RemoteTask.__init__(self, name=name, ssh_username=ssh_username, keypath=keypath, command=command, ip=ip,
                            working_dir=working_dir, globusendpoint=globusendpoint)
        self.docker_client = DockerRemoteClient(self.ssh_connection)

    def on_execute(self, launcher_script, script_name):
        """
        Execute the task script

        :param script: script content
        :type script: str

        :param script_name: script name
        :type script_name: str

        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)

        """

        RemoteTask.on_execute(self, launcher_script, script_name)
        return self.ssh_connection.execute_command("bash " + self.working_dir + "/.dagon/" + script_name)

    def on_garbage(self):
        """
        Call garbage collector, removing the scratch directory, containers and instances related to the
        task
        """
        RemoteTask.on_garbage(self)
        self.remove_container()



from globus_sdk import TransferData, AccessTokenAuthorizer, TransferClient
import globus_sdk
import os 

class GlobusManager:
    """
    **Manages the data transference between two nodes using globus**

    :ivar _from: endpoint ID of the data source machine
    :vartype _from: str

    :ivar _to: endpoint ID of the data destiny machine
    :vartype _to: str

    """

    TRANSFER_TOKEN = "Ag82ObxMdj80Gxd2ex7oVJWJnPv4dWQ8ndkpblOz51n42G6g96i8Cjqx5zKDE4jM5E2OzrJgMYMlNYC7NmMbYfk6VP"

    def __init__(self, _from, _to, client_id, intermediate):

        """

        :param _from: endpoint ID of the data source machine
        :type _from: str

        :param _to:  endpoint ID of the data destiny machine
        :type _to: str
        """

        self._from = _from
        self._to = _to
        self.intermediate = intermediate

        # Initialize the Globus Native App Client to transfer data
        #print(client_id)

        client = globus_sdk.NativeAppAuthClient(client_id)
        client.oauth2_start_flow()
        authorize_url = client.oauth2_get_authorize_url()
        print(f"Please go to this URL and login:\n\n{authorize_url}\n")

        auth_code = input("Please enter the code you get after login here: ").strip()
        token_response = client.oauth2_exchange_code_for_tokens(auth_code)

        globus_auth_data = token_response.by_resource_server["auth.globus.org"]
        globus_transfer_data = token_response.by_resource_server["transfer.api.globus.org"]

        globus_auth_token = globus_auth_data["access_token"]
        transfer_token = globus_transfer_data["access_token"]

        #print("AUTH GLOBUS TOKEN    ", globus_auth_token)
        #print("TRANSFER TOKEN    ", transfer_token)
        
        authorizer = globus_sdk.AccessTokenAuthorizer(transfer_token)
        self.transfer_client = globus_sdk.TransferClient(authorizer=authorizer)
        # Crea un cliente de transferencia de Globus
        #transfer_token = globus_sdk.RefreshTokenAuthorizer(globus_auth_token, client)
        #self.transfer_client = globus_sdk.TransferClient(authorizer=transfer_token)

    def copy_directory(self, ori, destiny, intermediate):

        """
        copy a directory using globus transfer

        :param ori: path where the data is in the source machine
        :type ori: str

        :param destiny: path where the data will be put on the destiny machine
        :type destiny: str

        :param tc: globus transfer client
        :type tc: :class:`globus_sdk.TransferClient`

        :return: status of the transference
        :rtype: str
        """

        print("ori: ", self._from)
        print("destiny: ", self._to)

        task_data = globus_sdk.TransferData(
            source_endpoint=self._from, destination_endpoint=self._to
        )

        #transference_data = TransferData(tc, self._from,
        #                                 self._to,
        #                                 label="SDK example",
        #                                 sync_level="checksum")

        task_data.add_item(ori, destiny, recursive=True)

        transfer_result = self.transfer_client.submit_transfer(task_data)
        #transfer_result = tc.submit_transfer(transference_data)
        while not self.transfer_client.task_wait(transfer_result["task_id"], timeout=1):
            task = self.transfer_client.get_task(transfer_result["task_id"])

            if task['nice_status'] == "NOT_A_DIRECTORY":
                self.transfer_client.cancel_task(task["task_id"])
                return task['nice_status']
        return "OK"

    def copy_file(self, ori, destiny, intermediate):
        """
        copy a file using globus

        :param ori: path where the data is in the source machine
        :type ori: str

        :param destiny: path where the data will be put on the destiny machine
        :type destiny: str

        :param tc: globus transfer client
        :type tc: :class:`globus_sdk.TransferClient`

        :return: status of the transference
        :rtype: str
        """

        # transference_data = TransferData(tc, self._from,
        #                                  self._to,
        #                                  label="SDK example",
        #                                  sync_level="checksum")

        #copy the data from the source endpoint to the Globus intermediate endpoint

        task_data = globus_sdk.TransferData(
            source_endpoint=self._from, destination_endpoint=self.intermediate
        )
        
        task_data.add_item(ori, intermediate, recursive=False)
        transfer_result = self.transfer_client.submit_transfer(task_data)
        
        #print(f"Transferencia iniciada. ID de tarea: {transfer_result['task_id']}")

        while not self.transfer_client.task_wait(transfer_result["task_id"], timeout=1):
            continue # wait until transfer ends

        #copy the data from the source endpoint to the Globus intermediate endpoint

        task_data = globus_sdk.TransferData(
            source_endpoint=self.intermediate, destination_endpoint=self._to
        )

        task_data.add_item(intermediate, destiny, recursive=False)
        transfer_result = self.transfer_client.submit_transfer(task_data)
        
        #print(f"Transferencia iniciada. ID de tarea: {transfer_result['task_id']}")

        while not self.transfer_client.task_wait(transfer_result["task_id"], timeout=1):
            continue # wait until transfer ends


        return "OK"

    def copy_data(self, ori, destiny, intermediate):

        """
        copy data using globus

        :param ori: path where the data is in the source machine
        :type ori: str

        :param destiny: path where the data will be put on the destiny machine
        :type destiny: str

        :raises Exception: a problem occurred during the transference
        """

        #authorizer = AccessTokenAuthorizer(GlobusManager.TRANSFER_TOKEN)
        #tc = TransferClient(authorizer=authorizer)

        if os.path.isdir(ori):  # if the path is a directory
            res = self.copy_directory(ori, destiny, intermediate)
        else:
            res = self.copy_file(ori, destiny, intermediate)

        if res is not "OK":
            raise Exception(res)


class SKYCDS:
    #TODO: Read this from configuration file
    CLIENT_TOKEN = "3c2d53762ec82cf4cb14f3c6d45601afaf4b2eb42c702fb9f9cc53fa874cf9a0"
    CATALOG_TOKEN = "b341db39dc182f276a4685ad4c0c8eb64bef1e7e1217655ff3da6eb28095670e"
    API_TOKEN = "16970b17feb38ad94a29443954487f8cde3221d2"
    IP_SKYCDS = ""

    def upload_data(self, task, path, mode="single", encryption=False):
        str_encryption = "true" if encryption else "false"
        command = "tar -czvf %s/data.tar %s --exclude=*.tar &&  docker exec -i client java -jar -Xmx3g -Xmx3g CP-ABE_ST_Up.jar %s %s %s %s bob 2 %s test %s" % \
                  (
                  task.get_scratch_dir(), path, SKYCDS.CLIENT_TOKEN, SKYCDS.API_TOKEN, SKYCDS.CATALOG_TOKEN, mode, path,
                  str_encryption)
        result = task.execute_command(command)
        return result

    def download_data(self, task, path):
        command = "mkdir -p %s && docker exec -i client java -jar -Xmx3g -Xmx3g CP-ABE_ST_Dow.jar %s %s %s %s 2 1 test %s" % \
                  (path, SKYCDS.CLIENT_TOKEN, SKYCDS.API_TOKEN, SKYCDS.CATALOG_TOKEN, SKYCDS.IP_SKYCDS, path)
        result = task.execute_command(command)
        return result



from flask import Flask, request
from flask import jsonify
from threading import Thread
from flask_api import status
from werkzeug.serving import make_server


class WorkflowServer(Thread):

    def __init__(self, workflow, ip, port):
        Thread.__init__(self)
        self.workflow = workflow
        self.ip = ip
        self.port = port
        self.configure_server()

    def run(self):
        self.srv.serve_forever()

    def shutdown(self):
        self.srv.shutdown()

    def configure_server(self):
        app = Flask(__name__)

        ######### FLASK FUNCTIONS ##################

        @app.route('/api/<task>/info', methods=['POST'])
        def info(task):
            if not request.is_json:
                return jsonify({"status": "error", "message": "Invalid JSON format"}), status.HTTP_400_BAD_REQUEST
            data = request.json
            task = self.workflow.find_task_by_name(self.workflow.name,task)
            if task is not None:
                task.set_info(data)
            return jsonify(data)

        @app.route('/check')
        def check():
            return jsonify({"status": "ok"})

        @app.route('/')
        def index():
            return "Hello world"

        ############################################

        self.srv = make_server(self.ip, self.port, app)
        self.ctx = app.app_context()
        self.ctx.push()



from requests.exceptions import ConnectionError

import requests
from requests.exceptions import MissingSchema


# Perform the communication with the dagon service
class API:
    def __init__(self, url):
        self.base_url = url
        self.checkConnection()

    # check if the service URL is valid or a service is available
    def checkConnection(self):
        """
        check if the service URL is valid or a service is available

        :raises ConnectionError: when it's not possible to connect to the URL provided
        """
        try:
            requests.head(self.base_url+"/check")
        except ConnectionError as e:
            raise ConnectionError("It is not possible connect to the URL %s" % self.base_url)
        except MissingSchema:
            raise ConnectionError("Bad URL %s" % self.base_url)

    # create workflow on dagon service
    def create_workflow(self, workflow):
        """
        create workflow on dagon service

        :param workflow: :class:`dagon.Workflow` object to save
        :type workflow: :class:`dagon.Workflow`

        :return: workflow id
        :rtype: int

        :raises Exception: when there is an error with the registration or the workflow name already exists
        """
        service = "/create"
        url = self.base_url + service
        data = workflow.as_json()
        res = requests.post(url, json=data)
        if res.status_code == 201:  # created
            json_reponse = res.json()
            return json_reponse['id']
        else:
            if res.status_code == 409:
                raise Exception("Workflow name already exists %d %s" % (res.status_code, res.reason))
            else:
                raise Exception("Workflow error registration %d %s" % (res.status_code, res.reason))

    # add task to workflow
    def add_task(self, workflow_id, task):
        """
        add task to workflow

        :param workflow_id: workflow id of the tasks
        :type workflow_id: int

        :param task: task to add to the workflow
        :type task: :class:`dagon.task.Task`

        :raises Exception: when there is an error with the call
        """
        service = "/add_task/%s" % workflow_id
        url = self.base_url + service
        data = task.as_json()
        res = requests.post(url, json=data)
        if res.status_code != 201 and res.status_code != 200:  # error
            raise Exception("Something went wrong %d %s" % (res.status_code, res.reason))

    # update a task status in the server
    def update_task_status(self, workflow_id, task, status):
        """
        update a task status in the server

        :param workflow_id: workflow id of the tasks
        :type workflow_id: int

        :param task: name of the task
        :type task: str

        :param status: task status
        :type status: str

        :raises Exception: when there is an error with the call
        """
        service = "/changestatus/%s/%s/%s" % (workflow_id, task, status)
        url = self.base_url + service
        res = requests.put(url)
        if res.status_code != 201 and res.status_code != 200:  # error
            raise Exception("Something went wrong %d %s" % (res.status_code, res.reason))

    # get a task from the server
    def get_task(self, workflow_id, task):
        """
        get a task from the server

        :param workflow_id: workflow id of the tasks
        :type workflow_id: int

        :param task: name of the task
        :type task: str

        :return: task
        :rtype: :class:`dagon.task.Task`

        :raises Exception: when there is an error with the call
        """

        service = "/%s/%s" % (workflow_id, task)
        url = self.base_url + service
        res = requests.get(url)
        if res.status_code != 201 and res.status_code != 200:  # error
            raise Exception("Something went wrong %d %s" % (res.status_code, res.reason))
        else:
            task = res.json()
            return task

    # get a task from the server
    def get_workflow_by_name(self, workflow_name):
        """
        get a workflow id from the server

        :param workflow_name: workflow name
        :type workflow_name: str

        :return: id
        :rtype: :str

        :raises Exception: when there is an error with the call
        """

        service = "/getworkflow/%s" % workflow_name
        url = self.base_url + service
        res = requests.get(url)
        if res.status_code != 201 and res.status_code != 200:  # error
            raise Exception("Something went wrong, no Transversal workflow %s founded, %d %s" % (workflow_name,res.status_code, res.reason))
        else:
            workflow_id = res.text
            return workflow_id

    # update atribute of the task
    def update_task(self, workflow_id, task, attribute, value):
        """
        update attribute of the task

        :param workflow_id: workflow id of the tasks
        :type workflow_id: int

        :param task: name of the task
        :type task: str

        :param attribute: attribute of the task to update
        :type task: str

        :param value: value of the attribute
        :type task: str

        :raises Exception: when there is an error with the call
        """

        service = "/update/%s/%s/%s?value=%s" % (workflow_id, task, attribute, value)
        url = self.base_url + service
        res = requests.put(url)
        if res.status_code != 201 and res.status_code != 200:  # error
            raise Exception("Something went wrong %d %s" % (res.status_code, res.reason))

    # add dependency on task
    def add_dependency(self, workflow_id, task, dependency):
        """
        add a dependency to an existing task in a workflow

        :param workflow_id: workflow id of the tasks
        :type workflow_id: int

        :param task: name of the task
        :type task: str

        :param dependency: name of the dependency
        :type task: str

        :raises Exception: when there is an error with the call
        """
        service = "/%s/%s/dependency/%s" % (workflow_id, task, dependency)
        url = self.base_url + service
        res = requests.put(url)
        if res.status_code != 201 and res.status_code != 200:  # error
            raise Exception("Something went wrong %d %s" % (res.status_code, res.reason))



import paramiko
from fabric.api import local
from paramiko import SSHClient

from dagon.communication import is_port_open


# To manage SSH connections
class SSHManager:

    """
    Manages SSH connections with a remote machine

    :ivar username: ssh username
    :vartype username: str

    :ivar host: IP of the remote machine
    :vartype host: str

    :ivar keypath: path to the private key
    :vartype keypath: str

    :ivar connection: connection with the remote machine
    :vartype connection: str

    """

    def __init__(self, username, host, keypath):

        """
        :param username: ssh username
        :type username: str

        :param host: IP of the remote machine
        :type host: str

        :param keypath: path to the private key
        :type keypath: str
        """

        self.username = username
        self.host = host
        self.keypath = keypath
        self.connection = self.get_ssh_connection()

    def get_connection(self):
        """
        return the connection

        :return: connection
        """
        return self.connection

    def create_file(self, filepath, content):
        """
        creates a file on the remote machine

        :param filepath: path to the file
        :type filepath: str

        :param content: content of the file
        :type filepath: str
        """

        sftp = self.get_connection().open_sftp()
        f = sftp.open(filepath, 'w')
        f.write(content)
        f.close()

    def get_ssh_connection(self):
        """
        returns a new ssh connection with the remote machine

        :return: ssh connection
        """

        ssh = SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        while not is_port_open(self.host, 22):
            continue
        if self.keypath is None:
            ssh.connect(self.host, username=self.username)
        else:
            ssh.connect(self.host, username=self.username, key_filename=self.keypath)
        return ssh

    def execute_command(self, command):
        """
        execute command in remothe machine over SSH

        :param command: command to execute on the remote machine
        :type command: str

        :return: execution results
        :rtype: dict(str, object)
        """

        _, stdout, stderr = self.connection.exec_command(command)
        code = stdout.channel.recv_exit_status()
        stdout = "\n".join(stdout.readlines())
        stderr = "\n".join(stderr.readlines())
        if len(stderr):
            return {"code": 1, "message": stderr}
        elif code > 0:
            return {"code": 1, "message": stdout}
        else:
            return {"code": 0, "output": stdout}


import shutil
import glob
from json import loads
from threading import Thread
from threading import Semaphore
from os import makedirs, path, chmod
from time import time, sleep
from enum import Enum
from dagon.ftp_publisher import FTP_API
import dagon


class TaskType(Enum):
    """
    Enum used to represent the main types tasks supported by dagon,
    remote tasks are supported by the RemoteClass but it should not be
    explicitly instantiate

    :cvar BATCH: Regular Batch task (local or remote)
    :cvar SLURM: Task executed using Slurm (local or remote)
    :cvar CLOUD: Task executed in a cloud instance (ec2, digital ocean and google cloud  tested with libcloud)
    :cvar DOCKER: Task executed on a docker container (local or remote)
    """

    BATCH = "batch"
    SLURM = "slurm"
    CLOUD = "cloud"
    DOCKER = "docker"


# Different types os tasks and their module and class name
tasks_types = {
    TaskType.BATCH: ("dagon.batch", "Batch"),
    TaskType.CLOUD: ("dagon.remote", "CloudTask"),
    TaskType.DOCKER: ("dagon.docker_task", "DockerTask"),
    TaskType.SLURM: ("dagon.batch", "Slurm")
}


class DagonTask(object):
    """
    Create the instance for Dagon Tasks
    """

    def __new__(cls, class_type, *args, **kwargs):
        """
        Choose the task type and returns an instance of the task type class

        :param class_type: type of the task (:class:`Types`)
        :type class_type: :class:`TaskType`

        :param args: arguments of the task
        :type name: list[]

        :param kwargs: keyboard arguments of the task
        :type name: dict()

        :return: subclass instance of :class:`Task`
        :rtype: cls
        """

        from importlib import import_module
        task_class = tasks_types[class_type]
        module = import_module(task_class[0])
        class_ = getattr(module, task_class[1])
        return class_(*args, **kwargs)


class Task(Thread):
    """
    **Represents a task executed by DagOn**

    It can be executed on local machine, cloud instance, cluster or in a docker container

    :ivar name: unique name of the class
    :vartype name: str

    :ivar command: command to be executed
    :vartype command: str

    :ivar working_dir: path to the folder or directory where the task is going to be executed
    :vartype working_dir: str

    :ivar nexts: tasks with dependencies to this tasks, that will be executed when this task ends
    :vartype nexts: list[]

    :ivar prevs: tasks that has to be executed before of this task (dependencies to be resolved)
    :vartype prevs: list[]

    :ivar reference_count: number of references to this task
    :vartype reference_count: int

    :ivar remove_scratch_dir: True if the sratch directory has to be removed after the execution of this task
    :vartype remove_scratch_dir: bool

    :ivar running: True if the task is in execution
    :vartype running: bool

    :ivar workflow: workflow related to this task
    :vartype workflow: :class:`dagon.Workflow`

    :ivar status: actual :class:`dagon.Status` of the task
    :vartype status: :class:`dagon.Status`

    :ivar info: information of the enviroment where this task is going to be executed
    :vartype info: dict(str, object)

    """

    def __init__(self, name, command, working_dir=None, transversal_workflow=None, globusendpoint=None):
        """
        :param name: name of the task
        :type name: str

        :param command: command to be executed by the task
        :type command: str

        :param working_dir: path to the directory where the task is be executed
        :type working_dir: str

        :param endpoint: UUID of the Globus Endpoint to store the data.
        :type endpoint: str

        """
        Thread.__init__(self)
        self.ssh_connection = None
        self.name = name
        self.nexts = []
        self.prevs = []
        self.reference_count = 0
        self.remove_scratch_dir = False
        self.ip = None
        self.running = False
        self.workflow = None
        self.set_status(dagon.Status.READY)
        self.working_dir = working_dir
        self.command = command
        self.info = None
        self.dag_tps = None
        self.transversal_workflow = transversal_workflow
        self.workflows = None
        self.data_mover = None
        self.stager_mover = None
        self.mode = "sequential"
        self.globusendpoint = globusendpoint
        #print("endpoint: ", endpoint)
        #print("endpoint: ", self.endpoint)

    def get_endpoint(self):
        return self.globusendpoint
    
    def set_endpoint(self, globusendpoint):
        self.globusendpoint = globusendpoint

    def set_mode(self, mode):
        self.mode = mode

    def get_mode(self):
        return self.mode

    def set_data_mover(self, data_mover):
        """
        Change the mode of the stager. The information is used by
        :class:`dagon.Stager` to decide the mode (COPY/LINK)

        :param data_mover: mode of the stager
        :type info: class:`dagon.DataMover`
        """
        self.data_mover = data_mover

    def set_stager_mover(self, stager_mover):
        """
        :param data_mover: mode of the stager
        :type info: class:`dagon.StagerMover`
        """
        self.stager_mover = stager_mover

    def set_info(self, info):
        """
        Change the information of the machine where the task is going to be executed. The information is used by
        :class:`dagon.Stager` to decide the data transfer protocol/application

        :param info: Machine info (ip, protocols, username)
        :type info: dict(str, object)
        """

        self.info = info

    def get_ip(self):
        """
        Returns the ip of the machine where the task is executed

        :return: IP address
        :rtype: str
        """
        from dagon.remote import CloudTask
       
        if isinstance(self, CloudTask):
            return self.info["public_ip"]
        else:
            return self.info["ip"] if self.ip == None else self.ip

    def get_info(self):
        """
        Returns the complete information of the machine where the task is executed

        :return: Machine information collected
        :rtype: dict(str, object)
        """
        return self.info

    def get_user(self):
        """
        Return the username who execute the task on the remote machine

        :return: Unix Username
        :rtype: str
        """

        return self.info["user"]

    def get_scratch_dir(self):
        """
        Returns the task's scratch directory as soon as it's available

        :return: Absolute path to the scratch directory
        :rtype: str
        """

        while self.working_dir is None and self.status is not dagon.Status.FAILED:
            sleep(1)
        return self.working_dir

    def get_scratch_name(self):
        """
        Generates a unique name for the task scratch directory name

        :return: Name of the scratch directory
        :rtype: str
        """
        millis = int(round(time() * 1000))
        return str(millis) + "-" + self.name

    def as_json(self):
        """"
        Generates the JSON representation of the task

        :return: JSON task representation
        :rtype: dict(str, object)
        """

        json_task = {"name": self.name, "status": self.status.name,
                     "working_dir": self.working_dir, "nexts": [], "prevs": [],
                     "command": self.command, "type": type(self).__name__.lower()}

        for t in self.nexts:
            json_task['nexts'].append(t.name)
        for t in self.prevs:
            json_task['prevs'].append(t.name)
        return json_task

    def set_workflow(self, workflow):
        """
        Set the workflow which execute this task

        :param workflow: :class:`dagon.Workflow` instance
        :type workflow: :class:`dagon.Workflow`
        """
        self.workflow = workflow

    def set_dag_tps(self, DAG_tps):
        """
        Set the DAG_tps workflow which execute this task

        :param  DAG_tps: :class:`dagon.dag_tps` instance
        :type  DAG_tps: :class:`dagon.dag_tps`
        """
        self.dag_tps = DAG_tps

    # Set the current status
    def set_status(self, status):
        """
        Set the status for the current task

        :param status: status of the task
        :type status: :class:`dagon.task.Status`
        """
        self.status = status
        if self.workflow is not None:
            self.workflow.logger.debug("%s: %s", self.name, self.status)
            if self.workflow.is_api_available:
                self.workflow.api.update_task_status(self.workflow.workflow_id, self.name, status.name)

    def execute_command(self, command):
        """"
        Executes a command
        :param command: command to be executed
        """
        pass

    def add_dependency_to(self, task):
        """
        Add a dependency to other task

        :param task: :class:`dagon.task.Task` instance dependency
        :type task: :class:`dagon.task.Task`
        """
        task.nexts.append(self)
        self.prevs.append(task)

        if self.workflow.is_api_available:  # add in the server
            self.workflow.api.add_dependency(self.workflow.workflow_id, self.name, task.name)

    def add_transversal_point(self, task):
        """
        Add a dependency from other task

        :param task: :class:`dagon.task.Task` instance dependency
        :type task: :class:`dagon.task.Task`
        """
        self.prevs.append(task)

        if self.workflow.is_api_available:  # add in the server
            self.workflow.api.add_dependency(self.workflow.workflow_id, self.name, task.name)

    # Increment the reference count
    def increment_reference_count(self):
        """
        Increments the reference count
        """
        self.reference_count = self.reference_count + 1

    # Call garbage collector (remove scratch directory, container, cloud instace, etc)
    # implemented by each task class
    def on_garbage(self):
        """
        Call garbage collector, removing the scratch directory, containers and instances related to the
        task
        """
        shutil.move(self.working_dir, self.working_dir + "-removed")

    # Decremet the reference count
    def decrement_reference_count(self):
        """
        Decremet the reference count. When the number of references is equals to zero, the garbage collector is called
        """
        self.reference_count = self.reference_count - 1

        # Check if the scratch directory must be removed
        if self.reference_count == 0 and self.remove_scratch_dir is True:
            # Call garbage collector (remove scratch directory, container, cloud instace, etc)
            self.on_garbage()
            # Perform some logging
            self.workflow.logger.debug("Removed %s", self.working_dir)

    def set_semaphore(self, sem):
        self.semaphore = sem

    # Method overrided
    def pre_run(self):
        """
        Resolve task dependencies
        For each workflow:// in the command string
        1) Extract the referenced task
        2) Add a reference in the referenced task

        """
        # Index of the starting position
        pos = 0
        # get workflows of dag_tps
        if self.dag_tps is not None:
            self.workflows = self.dag_tps.workflows

        # Forever unless no anymore dagon.Workflow.SCHEMA are present
        while True:
            # Get the position of the next dagon.Workflow.SCHEMA
            pos1 = self.command.find(dagon.Workflow.SCHEMA, pos)

            # Check if there is no dagon.Workflow.SCHEMA
            if pos1 == -1:
                # Exit the forever cycle
                break

            # Find the first occurrent of a whitespace (or if no occurrence means the end of the string)
            pos2 = self.command.find(" ", pos1)

            # Check if this is the last referenced argument
            if pos2 == -1:
                pos2 = len(self.command)

            # Extract the parameter string
            arg = self.command[pos1:pos2]

            # Remove the dagon.Workflow.SCHEMA label
            arg = arg.replace(dagon.Workflow.SCHEMA, "")

            # Split each argument in elements by the slash
            elements = arg.split("/")

            # Extract the referenced task's workflow name
            workflow_name = elements[0]

            # The task name is the first element
            task_name = elements[1]
            
            # Set the default workflow name if needed
            if workflow_name is None or workflow_name == "":
                workflow_name = self.workflow.name

            # Extract the reference task object

            if self.workflows is None:
                task = self.workflow.find_task_by_name(workflow_name, task_name)
            else:
                for wf in self.workflows:
                    task = wf.find_task_by_name(workflow_name, task_name)
                    if task is not None:
                        break

            # Check if the refernced task is consistent
            if task is not None:
                # Add the dependency to the task
                if workflow_name == self.workflow.name:
                    self.add_dependency_to(task)
                else:
                    self.add_transversal_point(task)
                # Add the reference from the task
                task.increment_reference_count()

            if task is None:  # if is None means that task is from another WF maybe in the dagon service
                #self.workflow.logger.debug("Adding transversal point")
                #self.workflow.logger.debug(workflow_name)
                #self.workflow.logger.debug(task)
                if self.workflow.is_api_available:
                    workflow_id = self.workflow.api.get_workflow_by_name(workflow_name)
                    transversal_task = self.workflow.api.get_task(workflow_id, task_name)[
                        'task']  # get the task from the external workflow
                    transversal_task = DagonTask(TaskType[transversal_task['type'].upper()], transversal_task['name'],
                                                 transversal_task['command'],
                                                 transversal_workflow=workflow_id,
                                                 working_dir=transversal_task['working_dir'])
                    self.add_transversal_point(transversal_task)
                else:
                    raise ConnectionError("Dagon service is not available")

            # Go to the next element
            pos = pos2

    # Pre process command
    def pre_process_command(self, command):
        """
        Preprocess the command resolving the dependencies with other tasks

        :param command:
        :type command: command to be executed by the task

        :return: command preprocessed
        :rtype: str
        """
        #print(self.workflow.cfg["globus"])
        stager = dagon.Stager(self.data_mover, self.stager_mover, self.workflow.cfg)

        # Initialize the script
        header = "#! /bin/bash\n"
        header = header + "# This is the DagOn launcher script\n\n"
        header = header + "code=0\n"
        # Add and execute the howim script

        context_script = header + "cd " + self.working_dir + "/.dagon\n"
        context_script += header + self.get_how_im_script() + "\n\n"

        result = self.on_execute(context_script, "context.sh")  # execute context script


        if result['code']:
            raise Exception(result['message'])
        self.set_info(loads(result['output']))

        

        ### start the creation of the launcher.sh script
        # Create the header
        header = header + "# Change the current directory to the working directory\n"
        header = header + "cd " + self.working_dir + "\n"
        header = header + "if [ $? -ne 0 ]; then code=1; fi \n\n"
        header = header + "# Start staging in\n\n"

        # Create the body
        body = command

        # Index of the starting position
        pos = 0

        # Forever unless no anymore dagon.Workflow.SCHEMA are present
        while True:

            # Get the position of the next dagon.Workflow.SCHEcdMA
            pos1 = command.find(dagon.Workflow.SCHEMA, pos)

            # Check if there is no dagon.Workflow.SCHEMA
            if pos1 == -1:
                # Exit the forever cycle
                break

            # Find the first occurrent of a whitespace (or if no occurrence means the end of the string)
            pos2 = command.find(" ", pos1)

            # Check if this is the last referenced argument
            if pos2 == -1:
                pos2 = len(command)

            # Extract the parameter string
            arg = command[pos1:pos2]

            # Remove the dagon.Workflow.SCHEMA label
            arg = arg.replace(dagon.Workflow.SCHEMA, "")

            # Split each argument in elements by the slash
            elements = arg.split("/")

            # Extract the referenced task's workflow name
            workflow_name = elements[0]

            # The task name is the first element
            task_name = elements[1]

            # Get the rest of the string as local path
            local_path = arg.replace(workflow_name + "/" + task_name, "")

            # Set the default workflow name if needed
            if workflow_name is None or workflow_name == "":
                workflow_name = self.workflow.name

            # Extract the reference task object
            # task = self.workflow.find_task_by_name(workflow_name, task_name)
            if self.workflows is None:
                task = self.workflow.find_task_by_name(workflow_name, task_name)
            else:
                for wf in self.workflows:
                    task = wf.find_task_by_name(workflow_name, task_name)
                    if task is not None: break

            if task is None:  # if is None means that task is from another WF maybe in the dagon service
                if self.workflow.is_api_available:
                    workflow_id = self.workflow.api.get_workflow_by_name(workflow_name)
                    response = self.workflow.api.get_task(workflow_id,
                                                          task_name)  # get the task from the external workflow
                    transversal_task = response['task']
                    host_ip = response['host']
                    # if the host is the same in this computer, the task is in the same computer
                    if host_ip == self.workflow.ftpAtt['host']:
                        task_path = transversal_task['working_dir']  # the same workingdir
                    else:
                        task_path = self.workflow.local_path + transversal_task[
                            'working_dir']  # if not, we add an extra path
                        ftp = FTP_API(host_ip)
                        task_folder = transversal_task['working_dir'].split("/")[-1]  # the last one is the task folder
                        ftp.downloadFiles(task_folder, task_path)
                        # we need to download the data from the ftp host

                    task = DagonTask(TaskType[transversal_task['type'].upper()], transversal_task['name'],
                                     transversal_task['command'],
                                     transversal_workflow=workflow_id, working_dir=task_path)

            # Check if the refernced task is consistent
            if task is not None:
                # Evaluate the destiation path
                dst_path = self.working_dir + "/.dagon/inputs/" + workflow_name + "/" + task_name

                # Create the destination directory
                header = header + "\n\n# Create the destination directory\n"
                header = header + "mkdir -p " + dst_path + "/" + path.dirname(local_path) + "\n"
                header = header + "if [ $? -ne 0 ]; then code=1; fi\n\n"
                # Add the move data command
                header = header + stager.stage_in(self, task, dst_path, local_path)

                if self.mode == "parallel":
                    files = glob.glob(task.get_scratch_dir() + "/" + local_path)
                    taskType = TaskType[type(self).__name__.upper()]

                    for file in files:
                        filename, _ = path.splitext(path.basename(file))
                        taskParallelName = "{}_{}".format(self.name, filename)
                        cmd = body.replace(dagon.Workflow.SCHEMA + arg, " workflow:///" + self.name + "/" + path.basename(file))

                        if type(self) == dagon.batch.Batch:
                            parallel_task = DagonTask(taskType, taskParallelName, cmd, 
                                                      transversal_workflow=self.transversal_workflow)
                        
                        if type(self) == dagon.batch.RemoteBatch:
                            parallel_task = DagonTask(taskType, taskParallelName, cmd, ssh_username=self.ssh_username, 
                                                      keypath=self.keypath, ip=self.ip)

                        elif type(self) == dagon.batch.Slurm:
                            parallel_task = DagonTask(taskType, taskParallelName, cmd, partition=self.partition, 
                                                      ntasks=self.ntasks, memory=self.memory)

                        elif type(self) == dagon.batch.RemoteSlurm:
                            parallel_task = DagonTask(taskType, taskParallelName, cmd, partition=self.partition, 
                                                      ntasks=self.ntasks, memory=self.memory,
                                                      ssh_username=self.ssh_username, keypath=self.keypath, ip=self.ip)

                        elif type(self) == dagon.remote.CloudTask:
                            parallel_task = DagonTask(taskType, taskParallelName, cmd, provider=self.provider, 
                                                      ssh_username=self.ssh_username, key_options=self.key_options, 
                                                      instance_id=self.instance_id, instance_flavour=self.instance_flavour, 
                                                      instance_name=self.instance_name, stop_instance=self.stop_instance)

                        elif type(self) == dagon.docker_task.DockerTask:
                            parallel_task = DagonTask(taskType, taskParallelName, cmd, image=self.image, 
                                                      container_id=self.container_id, remove=self.remove, 
                                                      volume=self.volume, transversal_workflow=self.transversal_workflow)

                        elif type(self) == dagon.docker_task.DockerRemoteTask:
                            parallel_task = DagonTask(taskType, taskParallelName, cmd, image=self.image, 
                                                      container_id=self.container_id, ssh_username=self.ssh_username, 
                                                      keypath=self.keypath, ip=self.ip,
                                                      remove=self.remove, volume=self.volume, 
                                                      transversal_workflow=self.transversal_workflow)
                        
                        self.workflow.add_task(parallel_task)

                    self.workflow.make_dependencies()

                    body = "echo \"Starting parallel tasks...\"\n"
                    body += "ln -sf " + dst_path + "/" + local_path + " " + self.get_scratch_dir()
                else:
                    # Change the body of the command
                    body = body.replace(dagon.Workflow.SCHEMA + arg, dst_path + "/" + local_path)
            pos = pos2

        # Invoke the command
        header = header + "\n\n# Invoke the command\n"
        header = header + self.include_command(body)
        header = header + "if [ $? -ne 0 ]; then code=1; fi"
        return header

    # process the command to execute
    def include_command(self, body):
        """
        Include the command to execute in the script body
        :param body: Script body
        :return: Script body with the command
        """
        return body + " |tee " + self.working_dir + "/.dagon/stdout.txt\n"

    # Post process the command
    def post_process_command(self, command):
        """
        Add some post process commands after the task execution
        :param command: Command to be executed
        :return: Command post-processed
        """
        footer = command + "\n\n"
        footer = footer + "# Perform post process\n"
        footer += "exit $code"
        return footer

    # Method to be overrided
    def on_execute(self, script, script_name):
        """
        Execute the task script

        :param script: script content
        :type script: str

        :param script_name: script name
        :type script_name: str

        :return: execution result
        :rtype: dict() with the execution output (str) and code (int)
        """
        # The launcher script name
        script_name = self.working_dir + "/.dagon/" + script_name
        # Create a temporary launcher script
        file = open(script_name, "w")
        file.write(script)
        file.flush()
        file.close()
        chmod(script_name, 0o744)

    # create path using mkdirs
    def mkdir_working_dir(self, path):
        """
        Make the working directory

        :param path: Path to the working directory
        :type path: str
        """
        makedirs(path,exist_ok=True)

    def create_working_dir(self):
        """
        Create the working directory
        """
        if self.working_dir is None:
            # Set a scratch directory as working directory
            self.working_dir = self.workflow.get_scratch_dir_base() + "/" + self.get_scratch_name()

            # Set to remove the scratch directory
            self.remove_scratch_dir = True

        # Create scratch directory
        self.mkdir_working_dir(self.working_dir + "/.dagon")
        self.workflow.logger.debug("%s: Scratch directory: %s", self.name, self.working_dir)
        if self.workflow.is_api_available:  # change scratch directory on server
            try:
                self.workflow.api.update_task(self.workflow.workflow_id, self.name, "working_dir", self.working_dir)
            except Exception as e:
                self.workflow.logger.error("%s: Error updating scratch directory on server %s", self.name, e)

    def remove_reference_workflow(self):
        """
        Remove the reference
        For each workflow:// in the command
        """
        # Remove the reference
        # For each workflow:// in the command

        # Index of the starting position
        pos = 0

        # Forever unless no anymore dagon.Workflow.SCHEMA are present
        while True:
            # Get the position of the next dagon.Workflow.SCHEMA
            pos1 = self.command.find(dagon.Workflow.SCHEMA, pos)

            # Check if there is no dagon.Workflow.SCHEMA
            if pos1 == -1:
                # Exit the forever cycle
                break

            # Find the first occurrent of a whitespace (or if no occurrence means the end of the string)
            pos2 = self.command.find(" ", pos1)

            # Check if this is the last referenced argument
            if pos2 == -1:
                pos2 = len(self.command)

            # Extract the parameter string
            arg = self.command[pos1:pos2]

            # Remove the dagon.Workflow.SCHEMA label
            arg = arg.replace(dagon.Workflow.SCHEMA, "")

            # Split each argument in elements by the slash
            elements = arg.split("/")

            # Extract the referenced task's workflow name
            workflow_name = elements[0]

            # The task name is the first element
            task_name = elements[1]

            # Set the default workflow name if needed
            if workflow_name is None or workflow_name == "":
                workflow_name = self.workflow.name

            # Extract the reference task object
            # task = self.workflow.find_task_by_name(workflow_name, task_name)
            if self.workflows is None:
                task = self.workflow.find_task_by_name(workflow_name, task_name)
            else:
                for wf in self.workflows:
                    task = wf.find_task_by_name(workflow_name, task_name)
                    if task is not None: break

            # Check if the refernced task is consistent
            if task is not None:
                # Remove the reference from the task
                task.decrement_reference_count()

            # Go to the next element
            pos = pos2

    # Method execute
    def execute(self):
        """
        Execute the task

        :raises Exception: a problem occurred during the task  execution
        """
        self.create_working_dir()

        # Apply some command pre processing
        launcher_script = self.pre_process_command(self.command)
        # Apply some command post processing
        launcher_script = self.post_process_command(launcher_script)

        # Execute only if not dry
        if self.workflow.dry is False:
            # Invoke the actual executor
            start_time = time()
            self.result = self.on_execute(launcher_script, "launcher.sh")
            self.workflow.logger.debug("%s Completed in %s seconds ---" % (self.name, (time() - start_time)))
            #print(self.result)
            # Check if the execution failed
            if self.result['code']:
                raise Exception('Executable raised a execption ' + self.result['message'])

        self.remove_reference_workflow()

    def run(self):
        """
        Runs the thread where the task will be executed
        """
        if self.workflow is not None:
            # Change the status

            self.set_status(dagon.Status.WAITING)

            # Wait for each previous tasks
            for task in self.prevs:
                if self.workflow.find_task_by_name(self.workflow.name, task.name) is None:  # if its a process from other workflow
                    while True:
                        if self.workflow.is_api_available and task.transversal_workflow is not None:  # when is an asynchronous execution
                            try:
                                transversal_task = self.workflow.api.get_task(task.transversal_workflow, task.name)[
                                    'task']  # get the task from the external workflow using the api
                                if transversal_task['status'] == dagon.Status.FINISHED.value or transversal_task[
                                    'status'] == dagon.Status.FAILED.value:
                                    break
                                else:
                                    sleep(1)
                            except Exception as e:
                                task.set_status(dagon.Status.FAILED)
                                self.workflow.logger.warning('Worflow dependence not found, Error: ' + str(e))
                                break
                        elif task.status == dagon.Status.FINISHED or task.status == dagon.Status.FAILED:  # when is used the dag_tps structure
                            break
                        else:
                            sleep(.5)
                else:
                    while True:
                        if task.status == dagon.Status.WAITING or task.status == dagon.Status.READY:  # if this happends, the workflow is probably a meta-workflow
                            sleep(.5)
                        else:
                            task.join()
                            break

            # Check if one of the previous tasks crashed
            for task in self.prevs:
                if task.status == dagon.Status.FAILED:
                    self.set_status(dagon.Status.FAILED)
                    return

            # Change the status
            self.set_status(dagon.Status.RUNNING)
            # Execute the task Job
            self.workflow.logger.debug("%s: Executing...", self.name)
            # self.semaphore.acquire()
            self.execute()
            sleep(2)
            # self.semaphore.release()
            """try:
                self.workflow.logger.debug("%s: Executing...", self.name)
                self.execute()
            except Exception, e:
                self.workflow.logger.error("%s: Except: %s", self.name, str(e))
                self.set_status(dagon.Status.FAILED)
                return
            # self.execute()"""

            # Start all next task
            for task in self.nexts:
                if task.status == dagon.Status.READY:
                    self.workflow.logger.debug("%s: Starting task: %s", self.name, task.name)
                    try:
                        task.start()
                    except:
                        self.workflow.logger.warn("%s: Task %s already started.", self.name, task.name)

            # Change the status
            # self.workflow.api.update_task(self.workflow.workflow_id, self.name, "working_dir", self.working_dir)
            self.set_status(dagon.Status.FINISHED)
            return

    def get_public_key(self):
        """
        Return the temporal public key to this machine

        :return: public key
        :rtype: str with the public key
        """
        pass

    def add_public_key(self, key):
        """
        Add a SSH public key on the remote machine

        :param key: Path to the public key
        :type key: str
        :return: result of the execution
        :rtype: dict() with the execution output (str) and code (int)
        """
        pass

    def remove_from_workflow(self):
        """
        Remove the reference to the workflow
        For each workflow:// in the command
        """
        # Remove the reference
        # For each workflow:// in the command

        # Index of the starting position
        pos = 0
        temp = self.command
        while True:
            # Get the position of the next dagon.Workflow.SCHEMA
            pos1 = self.command.find(dagon.Workflow.SCHEMA, pos)

            # Check if there is no dagon.Workflow.SCHEMA
            if pos1 == -1:
                # Exit the forever cycle
                break

            # Find the first occurrent of a whitespace (or if no occurrence means the end of the string)
            pos2 = self.command.find(" ", pos1)

            # Check if this is the last referenced argument
            if pos2 == -1:
                pos2 = len(self.command)

            # Extract the parameter string
            arg = self.command[pos1:pos2]

            # Remove the dagon.Workflow.SCHEMA label
            arg = arg.replace(dagon.Workflow.SCHEMA, "")

            # Split each argument in elements by the slash
            elements = arg.split("/")

            # Extract the referenced task's workflow name
            workflow_name = elements[0]

            # Set the default workflow name if needed
            if workflow_name is None or workflow_name == "":
                pass
            else:
                temp = temp.replace(workflow_name, "")

            # Go to the next element
            pos = pos2
        return temp

    def get_how_im_script(self):
        """
        Create the script to get the context where the task will be executed

        :return: Context script
        :rtype: str
        """
        return """
# Initialize
machine_type="none"
public_id="none"
user="none"
status_sshd="none"
status_ftpd="none"
status_skycds="none"

#get http communication protocol
curl_or_wget=$(if hash curl 2>/dev/null; then echo "curl"; elif hash wget 2>/dev/null; then echo "wget"; fi);


if [ $curl_or_wget = "wget" ]; then
  public_ip=`wget -q -O- https://ipinfo.io/ip`
else
  public_ip=`curl -s https://ipinfo.io/ip`
fi



# If no public ip is available, then it is a cluster node
machine_type="cluster-node"
private_ip=`ifconfig 2>/dev/null| grep "inet "| grep -v "127.0.0.1"| awk '{print $2}'|head -n 1`

if [ "$private_ip" == "" ]
then
  # The machine is a cluster frontend (or a single machine)
  machine_type="cluster-frontend"
  private_ip=`ifconfig 2>/dev/null| grep "inet "| grep -v "127.0.0.1"| awk '{print $2}'|grep -v "192.168."|grep -v "172.16."|grep -v "10."|head -n 1`
fi

#net-tools is not installed, try with ip -o route

if [ "$private_ip" == "" ]
then
  # The machine is a cluster frontend (or a single machine)
  machine_type="cluster-frontend"
  private_ip=`ip -o route get to 8.8.8.8 | sed -n 's/.*src \([0-9.]\+\).*/\\1/p'`
fi

# Check if the secure copy is available
status_sshd=`systemctl status sshd 2>/dev/null | grep 'Active' | awk '{print $2}'`
if [ "$status_sshd" == "" ]
then
  status_sshd="none"
fi

# Check if the ftp is available
status_ftpd=`systemctl status globus-gridftp-server 2>/dev/null|grep "Active"| awk '{print $2}'`
if [ "$status_ftpd" == "" ]
then
  status_ftpd="none"
fi

# Check if the grid ftp is available
status_gsiftpd=`systemctl status globus-gridftp-server 2>/dev/null|grep "Active"| awk '{print $2}'`
if [ "$status_gsiftpd" == "" ]
then
  status_gsiftpd="none"
fi

#check if skycds container is running
status_docker=`systemctl status docker 2>/dev/null|grep "Active"| awk '{print $2}'`
if [ "$status_gsiftpd" == "active" ]
then
    if [ "$(docker ps -aq -f status=running -f name=client)" ]; then
    # cleanup
        status_skycds="active"
    fi
fi

# Get the user
user=$USER

echo "no" | ssh-keygen  -b 2048 -t rsa -f ssh_key -q -N ""  >/dev/null

# Construct the json
json="{\\\"type\\\":\\\"$machine_type\\\",\\\"public_ip\\\":\\\"$public_ip\\\",\\\"ip\\\":\\\"$private_ip\\\",\\\"user\\\":\\\"$user\\\",\\\"SCP\\\":\\\"$status_sshd\\\",\\\"FTP\\\":\\\"$status_ftpd\\\",\\\"GRIDFTP\\\":\\\"$status_gsiftpd\\\",\\\"SKYCDS\\\":\\\"$status_skycds\\\"}"
echo $json
"""